"""枚举数据模型
"""

import enum
import typing
from typing import Optional as Opt, Annotated as Anno, Literal as Lit

from ..exceptions import InvalidStatusTransition


EnumClassTV = typing.TypeVar('EnumClassTV', bound=typing.Type[enum.Enum])
"""TypeVar of Enum class"""
EnumMemberTV = typing.TypeVar('EnumMemberTV', bound=enum.Enum)
"""TypeVar of Enum memebr(instance)
"""

@enum.unique
class Status(enum.Enum):

    """
    Examples
    --------
    .. code-block:: python
        from blue_firmament.scheme.enum import Status

        class MyStatus(Status):

            OPEN = "open"
            CLOSED = "closed"
            CANCELLED = "cancelled"

            def to_cancelled(self) -> typing.Literal["MyStatus.CANCELLED"]:
                return self.to_target_status(MyStatus.CANCELLED, 
                    MyStatus.OPEN  # break a line here to disguish allowed_from from target
                )

        MyStatus.CLOSED.to_cancelled()  
        # raise InvalidStatusTransition
        MyStatus.OPEN.to_cancelled() 
        # return MyStatus.CANCELLED
    """

    def _to_target_status(self, 
        target: EnumMemberTV,
        *allowed_from: "Status",
    ) -> EnumMemberTV:
        
        """Return target status if allowed (Idempotent)

        :param target: 
        :param allowed_from: 
            Allowed source status to go to the target status
        :raises InvalidStatusTransition: 
            When current status is not in `allowed_from`.
            Not include case that is already at target status.
        :return: The target status
        """
        if self is not target:
            if self not in allowed_from:
                raise InvalidStatusTransition.from_enum_member(self, target)
        return target



"""Common Task Context
"""

__all__ = [
    "CommonTaskContext",
    "SoCommonTC"
]

import typing
from ... import event
from ...scheme import FieldT, private_field
from ..context import SoBaseTC, ExtendedTaskContext
from ...session.common import CommonSession
if typing.TYPE_CHECKING:
    from ...dal import DataAccessObjects


class CommonTaskContext(
    ExtendedTaskContext[CommonSession],
    session_cls=CommonSession
):
    """Task context extended with common session.

    .. versionadded:: 0.1.2
    """

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__()

    @property
    def _emit(self):
        return event.simple_emit
    @property
    def _daos(self):
        return self._session.daos
    @property
    def _operator(self):
        return self._session.operator


class SoCommonTC(SoBaseTC):

    _task_context: FieldT[CommonTaskContext] = private_field(
        default_factory=CommonTaskContext.from_contextvar
    )

    @property
    def _operator(self): return self._task_context._operator
    @property
    def _daos(self): return self._task_context._daos
    @property
    def _dao(self): return self._task_context._daos(self.__class__)



__all__ = [
    'singleton',
    'call_as_async',
    'call_as_sync',
    'dump_iterable',
    'try_convert_str',
]

import asyncio
import inspect
import threading
import typing
import nest_asyncio

T = typing.TypeVar('T')


def singleton(cls):

    """
    A decorator to implement a global single instance for a class.
    This decorator uses a lock to ensure thread-safe instantiation.
    """
    _instances = {}  # Store instances of decorated classes
    _lock = threading.Lock()

    def get_instance(*args, **kwargs):
        """
        Retrieves or creates the singleton instance.
        """
        with _lock:  # Ensure thread-safe instantiation
            if cls not in _instances:
                _instances[cls] = cls(*args, **kwargs)

        return _instances[cls]

    return get_instance


async def call_as_async(
    func: typing.Callable, *args, **kwargs
) -> typing.Any:
    """Call both sync and async function asynchronously.

    :param func: The async/sync function to call
    :param args: Positional arguments to pass to the function
    :param kwargs: Keyword arguments to pass to the function

    Returns
    --------
    The result of the function call

    Example
    --------
    ```python
        def sync_func(x):
            return x * 2

        async def async_func(x):
            await asyncio.sleep(0.1)
            return x * 3

        # Both can be called the same way
        result1 = await call_function_as_async(sync_func, 5)  # Returns 10
        result2 = await call_function_as_async(async_func, 5)  # Returns 15
    ```
    """
    call = getattr(func, "__call__", None)

    if (
        call is not None and inspect.iscoroutinefunction(call)
        or
        inspect.iscoroutinefunction(func)
    ):
        # Function is async, run it with await
        return await func(*args, **kwargs)
    else:
        # Function is synchronous, call it directly
        return func(*args, **kwargs)


def call_as_sync(func: typing.Callable, *args, **kwargs) -> typing.Any:
    """Call both sync and async function synchronously.
    """
    res = func(*args, **kwargs)
    if isinstance(res, typing.Coroutine):
        nest_asyncio.apply()
        return asyncio.run(res)
    return res


IterT = typing.TypeVar('IterT', bound=typing.Iterable)
def dump_iterable(seq_cls: type[IterT], value: typing.Any) -> IterT:
    if isinstance(value, seq_cls):
        return value
    else:
        return seq_cls(value)


def try_convert_str(value: str) -> typing.Union[str, int, float, bool, None]:
    """
    Attempts to convert a string value to a Python primitive type.
    """
    if value.isdigit():
        return int(value)
    try:
        return float(value)
    except ValueError:
        pass
    if value.lower() == 'true':
        return True
    elif value.lower() == 'false':
        return False
    if value.lower() == 'null' or value == '':
        return None
    return value



'''CommonManager Module
'''

__all__ = [
    "CommonManager", 
    "PresetHandlerConfig",
]

from dataclasses import dataclass
import typing
from typing import Literal as Lit, Optional as Opt, Annotated as Anno

from ..scheme import EditableScheme
from .. import event
from ..utils.exec_ import build_func_sig
from ..task.registry import TaskRegistry
from ..task.context.common import CommonTaskContext
from ..dal import KeyableType, DataAccessObject
from ..scheme.field import CompositeField, FieldValueProxy
from ..log.main import get_logger
# from .base import BaseFieldManager, 
from .base import BaseManager, SchemeTV
from ..utils.typing_ import safe_issubclass
from ..task.main import Method
from ..scheme import BaseScheme
from ..task import TaskID, TaskMetadata

if typing.TYPE_CHECKING:
    from ..core.app import BlueFirmamentApp
    from ..scheme.field import Field


logger = get_logger(__name__)


@dataclass
class PresetHandlerConfig:
    """Config what and how preset handlers added to a manager.

    - These handlers will be automatically added to manager router.
      (which means their task_id's path are prefixed with the
      path_prefix you configured from class init params.)
    - You can call these handlers but without a type hinting.

    """

    editable: Opt[typing.Type[BaseScheme]] = None
    """Editable ver of managing scheme.

    This scheme is used to control which fields are editable
    for user. 
    
    Like in create, put and patch, forbids user from changing
    fields like id, created_at, but allows user modifying title,
    description...
    """
    sup_path: Opt[str] = None
    """Path append to the manager path prefix.
    
    If provided, handler's key param naming will be the same of this.
    """
    key_fields: Opt[dict[str, "Field"]] = None
    """Map key name to its field instance.

    For scheme only has one key, usually is not required.
    But for scheme has two or more (composite) keys, required.
    """
    get: bool = False
    """Add handler getting managing scheme.

    - Name: ``get_<manager_name>``
    - TaskID: ``GET /{<manager_name>_id}``
    """
    get_a_field: typing.Iterable["Field"] = ()
    """Add field getter handler for these fields.

    - Name: `get_<manager_name>_<field.name>`
    - TaskID: ``GET /{<manager_name>_id}/<field.name>``
    """
    create: bool = False
    """Add create handler.

    - Requiring editable
    """
    # TODO ad put_as_patch
    put: bool = False
    """
    
    - Requiring editable
    """
    put_a_field: typing.Iterable["Field"] = ()
    insert_item: typing.Iterable["Field"] = ()
    patch: bool = False
    """
    
    - Requiring editable
    """
    delete: bool = False
    delete_item: typing.Iterable["Field"] = ()


TV = typing.TypeVar('TV')
KeyTV = typing.TypeVar('KeyTV', bound=KeyableType)
class CommonManager(
    typing.Generic[SchemeTV, KeyTV],
    BaseManager[SchemeTV],
    CommonTaskContext,
):
    """
    Configuration
    -------------
    - scheme_cls: Scheme class this manager is managing
    - path_prefix: Path prefix of this manager's router.
    - manager_name: Friendly name of this manager.
        - use snake_case
    - preset_handler_config:
    """

    def __init__(self, task_context: CommonTaskContext):
        BaseManager.__init__(self, task_context)
        CommonTaskContext.__init__(self, tc=task_context, skip_btc_init=True)

    def __init_subclass__(
        cls,
        preset_handler_config: Opt[PresetHandlerConfig] = None,
        **kwargs
    ):
        super().__init_subclass__(**kwargs)
        scheme_cls = cls.__scheme_cls__
        manager_name = cls.__manager_name__

        # REFACTOR
        if preset_handler_config:
            if not scheme_cls:
                raise ValueError("scheme cls is required if you want to set up \
                    preset handler")
            
            exec_namespaces = globals().copy()
            handlers: dict[str, typing.Callable] = {}

            exec_namespaces["Editable"] = preset_handler_config.editable

            key_field = scheme_cls._get_key_field()
            if isinstance(key_field, CompositeField) \
                and not preset_handler_config.key_fields:
                    raise ValueError("key fields required for composite key")

            
            key_aliases: typing.Sequence[str]
            if preset_handler_config.sup_path and preset_handler_config.key_fields:
                key_aliases = TaskID.resolve_dynamic_indices(
                    cls.__path_prefix__ + preset_handler_config.sup_path
                )
                sup_path = preset_handler_config.sup_path
                exec_namespaces.update({
                    f'{i}_conv': j.converter
                    for i, j in preset_handler_config.key_fields.items()
                })
            else:
                key_aliases = (f"{manager_name}_id",)
                sup_path = f"/{{{manager_name}_id}}"
                exec_namespaces[f"{key_aliases[0]}_conv"] = key_field.converter

            if preset_handler_config.get:
                handler_name = f'get_{manager_name}'
                func_sig = build_func_sig(
                    handler_name,
                    *(
                        (i, f"Anno[typing.Any, {i}_conv]")
                        for i in key_aliases
                    ),
                    async_=True,
                    method=True
                )
                if isinstance(key_field, CompositeField):
                    func_body = f"    return await self.get(_id={key_aliases[0]}_conv({ \
                        ",".join(
                            f"{preset_handler_config.key_fields[i].in_scheme_name}={i}" # type: ignore
                            for i in key_aliases
                        ) \
                    }))"
                else:
                    func_body = f"    return await self.get(_id={key_aliases[0]})"
                
                exec(func_sig + func_body, exec_namespaces, handlers)
                setattr(cls, handler_name, handlers[handler_name])

                cls.__task_registries__.setdefault("default", TaskRegistry(
                    name="default", path_prefix=cls.__path_prefix__
                )).add_handler(
                    method=Method.GET, path=sup_path,
                    function=handlers[handler_name],
                    handler_manager_cls=cls
                )

            if preset_handler_config.put:
                handler_name = f'put_{manager_name}'
                func_sig = build_func_sig(
                    handler_name,
                    ("body", "Editable"),
                    *(
                        (i, f"Anno[typing.Any, {i}_conv]")
                        for i in key_aliases
                    ),
                    async_=True,
                    method=True
                )
                if isinstance(key_field, CompositeField):
                    func_body = f"    return await self.put(editable=body, _id={key_aliases[0]}_conv({ \
                        ",".join(
                            f"{preset_handler_config.key_fields[i].in_scheme_name}={i}"  # type: ignore
                            for i in key_aliases
                        ) \
                    }))"
                else:
                    func_body = f"    return await self.put(editable=body, _id={key_aliases[0]})"

                exec(func_sig + func_body, exec_namespaces, handlers)
                setattr(cls, handler_name, handlers[handler_name])

                cls.__task_registries__.setdefault("default", TaskRegistry(
                    name="default", path_prefix=cls.__path_prefix__
                )).add_handler(
                    method=Method.PUT, path=sup_path,
                    function=handlers[handler_name],
                    handler_manager_cls=cls
                )

    @property
    def _dao(self) -> DataAccessObject[SchemeTV]:
        """DAO of managing scheme.
        """
        return self._daos(self._scheme_cls)

    def _emit(
        self,
        name: str,
        parameters: Opt[dict] = None,
        metadata: Opt[dict | TaskMetadata] = None,
        without_prefix: bool = False
    ):
        """:meth:`event.simple_emit` but prefix name with manager path prefix.

        :param name: Name of event. Starts with dot.
        :param metadata: TaskMetadata.
            If not provided, use current task's metadata.
        :param without_prefix:
            If True, do not prefix name with manager path prefix.
        """
        return event.simple_emit(
            name=f"{self.__path_prefix__.replace('/', '.') if not without_prefix else ""}{name}",
            parameters=parameters,
            metadata=metadata or self._task.metadata
        )

    async def _get_scheme(self, _id: Opt[KeyTV] = None) -> SchemeTV:
        """Get managing scheme.

        :param _id: Key value

            If not None, return a scheme which key matches _id value;
            otherwise return the managing scheme.

        :raise ValueError: no managing scheme and _id is None

        Examples
        --------
        .. code-block:: python
            class MyManager(CommonManager[My, KeyType]):
                ...
                async def my_handler(self, _id: Opt[IdType] = None):
                    my = await self._get_scheme(_id)
        """
        try:
            scheme = self._scheme
            if _id is not None:
                if scheme.key_value == _id:
                    return scheme
                raise ValueError
            return scheme
        except ValueError as e:
            if _id is not None:
                return await self.get(_id=_id)
            raise e

    async def get(self, _id: KeyTV) -> SchemeTV:
        """Get scheme

        If success, set as managing scheme.
        """
        self._scheme = await self._dao.select_one(
            _id, task_context=self
        )
        return self._scheme

    async def insert(self, scheme: Opt[SchemeTV] = None) -> SchemeTV:
        """插入数据模型实例到 DAO

        - 插入成功则设置为当前实例

        :param scheme: 数据模型实例；不提供则为当前实例

        """
        self._scheme = await self._dao.insert(
            to_insert=scheme or await self._get_scheme(),
        )
        return self._scheme

    async def patch(self, editable: EditableScheme, _id: Opt[KeyTV] = None) -> SchemeTV:
        """Patch managing scheme to DAO.

        :param editable: Editable version of managing scheme.
        :param _id: which scheme to put, if not provided, use current managing scheme.
        :return:
        """
        self._scheme = await self._get_scheme(_id=_id)
        self._scheme._merge(scheme=editable)
        return await self._update_scheme(self._scheme)
    
    async def _update_scheme(
        self,
        scheme: Opt[SchemeTV] = None,
    ) -> SchemeTV:
        """Update dirty fields to dal.

        If success, set update result to manager scheme.
        """
        self._scheme = await self._dao.update(
            to_update=scheme or await self._get_scheme(), 
            only_dirty=True,
        )
        return self._scheme

    async def get_a_field(
        self,
        field: "Field[TV]", 
        _id: Opt[KeyTV] = None,
    ) -> TV:
        """Get a field's value of managing scheme.

        :returns: Field Value (no proxy)
        """
        scheme = self._try_get_scheme()
        if not scheme:
            return await self._dao.select_a_field(
                field,
                self._scheme_cls._get_key_field().equals(_id),
                task_context=self
            )
        else:
            return FieldValueProxy.dump(scheme._get_value(field))
        
    async def put_a_field(
        self,
        field: "Field[TV]", 
        value: TV,  
        _id: Opt[KeyTV] = None,
    ) -> TV:
        
        """Put a field value of managing scheme.

        Put value will be performed on scheme so that
        related validators can be run, and then
        patch to dal.
        """        
        scheme = await self._get_scheme(_id=_id)
        scheme[field] = value
        return (await self._update_scheme(scheme))[field]

    IoDableT = typing.TypeVar('IoDableT', bound=typing.List | typing.Set)
    async def insert_item(
        self,
        field: "Field[IoDableT]",
        values: typing.Iterable[TV],
        _id: Opt[KeyTV] = None,
        mode: Lit['append', 'prepend', 'insert'] = 'append',
        at: Opt[int] = None,
    ) -> IoDableT:
        
        """插入条目到列表型字段值

        :param _id: 主键值；没有则从当前实例获取
        :param field: 要插入到的字段；类型必须是列表或集合
        :param values: 要插入的值（可多个）；
        :param mode: 插入模式 （Set 不生效）

            - append: `[raw][values]`
            - prepend: `[values][raw]`
            - insert: `[raw before at][values][raw after at]`
        :param at: 插入位置；仅当 mode 为 `insert`
        """
        field_value = await self.get_a_field(_id=_id, field=field) 

        if isinstance(field_value, list):
            if mode == 'append':
                field_value.extend(values)
            elif mode == 'prepend':
                field_value[:0] = values
            elif mode == 'insert':
                if at is None:
                    raise ValueError('at must be provided when mode is insert')
                field_value[at:at] = values
            else:
                raise ValueError(f'Invalid mode {mode} for insert item')
        elif isinstance(field_value, set):
            field_value.update(values)
        else:
            raise ValueError(f'DAO not returning valid value of field {field}, {field_value}')
        
        return await self.put_a_field(
            field=field, value=field_value, 
            _id=_id,
        )
    
    async def delete_item(
        self,
        field: "Field[IoDableT]", 
        values: typing.Union[
            typing.Iterable[TV],
            typing.Set[TV]
        ],
        _id: Opt[KeyTV] = None,
    ) -> IoDableT:
        """从列表型字段值中删除条目

        :param _id: 主键值；没有则从当前实例获取
        :param field: 要删除的字段；值必须是列表类型
        :param values: 要删除的值（可多个）；
        :param dao: 数据访问对象；不提供则使用当前会话的 DAO

        Implementation
        --------------
        高效删除
        ^^^^^^^^
        因为要删除的数量不可能大于当前列表的长度
        """
        field_value = await self.get_a_field(_id=_id, field=field)

        if isinstance(values, (list,)):
            field_value = [v for v in field_value if v not in values]
        elif isinstance(values, (set,)):
            for value in values:
                field_value.remove(value)
        else:
            raise ValueError(f'{field} value invalid, value is {field_value}')

        return await self.put_a_field(
            _id=_id, field=field, value=field_value
        )
    
    async def delete(self, _id: Opt[KeyTV] = None) -> None:
        """Delete the scheme from DAO

        If success, set managing scheme to None.

        :param _id: key value

            If not provided, use managing scheme's.
        """
        scheme = await self._get_scheme(_id=_id)
        await self._dao.delete(to_delete=scheme)
        self._scheme = None


# CommonManagerTV = typing.TypeVar('CommonManagerTV', bound=CommonManager)
# class CommonFieldManager(
#     typing.Generic[TV, KeyTV, CommonManagerTV],
#     BaseFieldManager[TV, CommonManagerTV]
# ):
    
#     __scheme_manager_cls__: typing.Type[CommonManagerTV]

#     async def get_field(self,
#         _id: Opt[KeyTV] = None
#     ) -> TV:
        
#         """获取字段值
#         """
#         if self._field_value is None:
#             scheme = await self.scheme_manager._get_scheme(_id=_id)
#             self._field_value = scheme[self.field]
        
#         return typing.cast(TV, self._field_value)
    
#     async def update_field(self,
#         _id: Opt[KeyTV] = None,
#         value: TV | Undefined = _undefined
#     ) -> TV:
        
#         """更新字段值
#         """
#         if value is _undefined:
#             value_ = self._field_value
#         else:
#             value_ = value
        
#         scheme = await self.scheme_manager._get_scheme(_id=_id)
#         scheme[self.field] = value_
#         return await self._mdao.update(scheme)

class GetAFieldOptions(typing.TypedDict):
    fields: typing.Iterable["Field"]

class CreateOptions(typing.TypedDict):
    disabled_fields: typing.Iterable["Field"]

class InsertItemOptions(typing.TypedDict):
    fields: typing.Iterable["Field"]

class DeleteItemOptions(typing.TypedDict):
    fields: typing.Iterable["Field"]

class PatchOptions(typing.TypedDict):
    available_fields: typing.Iterable["Field"]

class PutAFieldOptions(typing.TypedDict):
    fields: typing.Iterable["Field"]

class PutOptions(typing.TypedDict):
    disabled_fields: typing.Iterable["Field"]


ManagerType = typing.TypeVar('ManagerType', bound=CommonManager)
def common_handler_adder(
    path_prefix_with_key: Opt[str] = None,
    get: bool = False,  # TODO 添加鉴权器（防御式）
    get_a_field: bool = False,
    get_a_field_options: Opt[GetAFieldOptions] = None,
    create: bool = False,
    create_options: Opt[CreateOptions] = None,
    insert_item: bool = False,
    insert_item_options: Opt[InsertItemOptions] = None,
    patch: bool = False,
    patch_options: Opt[PatchOptions] = None,
    put_a_field: bool = False,
    put_a_field_options: Opt[PutAFieldOptions] = None,
    put: bool = False,
    put_options: Opt[PutOptions] = None,
    delete_item: bool = False,
    delete_item_options: Opt[DeleteItemOptions] = None,
    app: Opt["BlueFirmamentApp"] = None
):

    """A decorator adding common handlers to a manager.

    :param app: if provided, register handlers to app

    TODO
    ----
    - insert and delete item limit body's type more accurately (item's type)
        - this will bring a better validation before reach the handler
    - create
        - scheme provide a method to fastly reset disabled fields to default value
    - patch
    - put

    Features
    --------
    Get
    ^^^
    - A handler return a scheme using session DAO to select record by primary key

    Get a Field
    ^^^^^^^^^^^
    - A handler return a field of the scheme using session DAO to select record by primary key
    - Route `GET /<manager_name>/{<manager_name>_id}/<field_name>`

    Create
    ^^^^^^^^^^^^^^
    - A handler insert a record into session DAO with provided scheme data.
    - Only fields not in `disabled_fields` will be inserted, others will be ignored.

    Insert item
    ^^^^^^^^^^^^^^
    - A handler insert item(s) into an insertable field of the scheme using session DAO.
    - Route `POST /<manager_name>/{<manager_name>_id>/<field_name>`
    - Request body is a list of items to insert

    Patch
    ^^^^^
    - A handler update provided fields of the scheme into session DAO.
    - Only fields in `available_fields` will be updated, other will be ignored.

    Put a Field
    ^^^^^^^^^^^
    - A handler update a field of the scheme into session DAO.
    - Every field in `fields` will get a handler.
    - Route `PUT /<manager_name>/{<manager_name>_id}/<field_name>`
    - Request body is the new value to put

    Example
    -------
    .. code-block:: python
        @common_manager_adder(
            'user',
            put_a_field=True,
            put_a_field_options={
                'fields': (UserScheme.nickname,)
            }
        )
        class UserManager(BaseManager[UserScheme, Session]):
            __SCHEME_CLS__ = UserScheme
            __name__ = 'user'

        reg = UserManager.get_route_record_register(app)
        reg(GET, '')

    Implementation
    --------------
    Add handler to manager
    ^^^^^^^^^^^^^^^^^^^^^^^
    handler func are dynamically generated from `exec`, so it can has \n
    dynamic name
    """

    def wrapper(manager_cls: typing.Type[ManagerType]):

        manager_name = manager_cls.__manager_name__
        primary_key_name = manager_name + "_id"
        if path_prefix_with_key is None:
            path_prefix_with_key_ = f"{manager_name}/{{{primary_key_name}}}"
        else:
            path_prefix_with_key_ = path_prefix_with_key

        if app:
            register = manager_cls.get_route_register(app)
        else:
            register = None

        if safe_issubclass(manager_cls.__scheme_cls__, BaseScheme):
            primary_key_field = manager_cls.__scheme_cls__._get_key_field()
        else:
            primary_key_field = None

        result_namespaces = {}

        if get:
            if not primary_key_field:
                raise ValueError('primary key is required for get')
            
            exec_namespaces = globals().copy()
            
            get_handler_name = f'get_{manager_name}'
            func_sig = f"async def {get_handler_name}(self, \n"
            if isinstance(primary_key_field, CompositeField):
                exec_namespaces[f'{primary_key_name}'] = primary_key_field.vtype
                for i in primary_key_field.sub_fields:
                    exec_namespaces[f"{i.name}_validator"] = i.converter
                    func_sig += f"{i.name}: typing.Annotated[typing.Any, {i.name}_validator],"
                func_body = f"    return await self.get(_id={primary_key_name}({
                    f",".join(
                        f"{i.name}={i.name}"
                        for i in primary_key_field.sub_fields
                    )
                }))"
            else:
                exec_namespaces["pk_validator"] = primary_key_field.converter
                func_sig += f"{primary_key_name}: typing.Annotated[typing.Any, pk_validator]"
                func_body = f"    return await self.get(_id={primary_key_name})"
            func_sig += "):\n"
            func_str = func_sig + func_body
            
            exec(func_str, exec_namespaces, result_namespaces)
            setattr(
                manager_cls, get_handler_name, 
                result_namespaces[get_handler_name]
            )

            if app:
                app.task_registry.add_handler(
                    method=Method.GET,
                    path=path_prefix_with_key_,
                    handler=getattr(manager_cls, get_handler_name),
                    handler_manager_cls=manager_cls
                )

        if put_a_field:
            if not put_a_field_options:
                raise ValueError('options are required for put_a_field')
            if not primary_key_field:
                raise ValueError('primary key is required for put_a_field')
            
            exec_namespaces = globals().copy()
            exec_namespaces["pk_validator"] = primary_key_field.converter
            
            fields = put_a_field_options.get('fields')
            for field in fields:

                put_a_field_handler_name = f'put_{field.name}'
                primary_key_name = f"{manager_name}_id"
                func_sig = f"async def {put_a_field_handler_name}(self, body, \n"
                func_sig += f"{primary_key_name}: typing.Annotated[typing.Any, pk_validator]"
                func_sig += "):\n"
                func_body = f"    return await self.put_a_field(_id={primary_key_name},\n"
                func_body += f"        field=self.scheme_cls.{field.in_scheme_name},\n"
                func_body += f"        value=body\n"
                func_body += f"    ) \n"
                func_str = func_sig + func_body
                
                exec(func_str, exec_namespaces, result_namespaces)
                setattr(
                    manager_cls, put_a_field_handler_name, 
                    result_namespaces[put_a_field_handler_name]
                )

                if register:
                    register(
                        Method.PUT,
                        '/{' + primary_key_name + '}/' + field.name,
                        getattr(manager_cls, put_a_field_handler_name),
                    )

        if get_a_field:
            if not get_a_field_options:
                raise ValueError('options are required for get_a_field')
            if not primary_key_field:
                raise ValueError('primary key is required for get_a_field')
            
            exec_namespaces = globals().copy()
            exec_namespaces["pk_validator"] = primary_key_field.converter
            
            fields = get_a_field_options.get('fields')
            for field in fields:

                get_a_field_handler_name = f'get_{field.name}'
                primary_key_name = f"{manager_name}_id"
                func_sig = f"async def {get_a_field_handler_name}(self, \n"
                func_sig += f"    {primary_key_name}: typing.Annotated[typing.Any, pk_validator]"
                func_sig += "):\n"
                func_body = f"    return await self.get_a_field(_id={primary_key_name}, \n"
                func_body += f"        field=self.scheme_cls.{field.in_scheme_name}\n"
                func_body += "    )\n"
                func_str = func_sig + func_body
                
                exec(func_str, exec_namespaces, result_namespaces)
                setattr(
                    manager_cls, get_a_field_handler_name, 
                    result_namespaces[get_a_field_handler_name]
                )

                if register:
                    register(
                        Method.GET,
                        '/{' + primary_key_name + '}/' + field.name,
                        getattr(manager_cls, get_a_field_handler_name),
                    )

        if insert_item:
            if not insert_item_options:
                raise ValueError('options are required for insert_item')
            if not primary_key_field:
                raise ValueError('primary key is required for insert_item')
            
            exec_namespaces = globals().copy()
            exec_namespaces["pk_validator"] = primary_key_field.converter
            
            fields = insert_item_options.get('fields')
            for field in fields:
                insert_item_handler_name = f'insert_{field.name}'
                primary_key_name = f"{manager_name}_id"
                func_sig = f"async def {insert_item_handler_name}(self, body: typing.Sequence, \n"
                func_sig += f"{primary_key_name}: typing.Annotated[typing.Any, pk_validator],\n"
                func_sig += "at: typing.Optional[int] = None, mode: str = 'append',\n"
                func_sig += "):\n"
                func_body = f"    return await self.insert_item(_id={primary_key_name}, \n"
                func_body += f"        field=self.scheme_cls.{field.in_scheme_name},\n"
                func_body += "        mode=mode, at=at,\n"
                func_body += "        values=body\n"
                func_body += "    )\n"
                func_str = func_sig + func_body
                
                exec(func_str, exec_namespaces, result_namespaces)
                setattr(
                    manager_cls, insert_item_handler_name, 
                    result_namespaces[insert_item_handler_name]
                )

                if register:
                    register(
                        Method.POST,
                        '/{' + primary_key_name + '}/' + field.name,
                        getattr(manager_cls, insert_item_handler_name),
                    )
        
        if delete_item:
            if not delete_item_options:
                raise ValueError('options are required for delete_item')
            if not primary_key_field:
                raise ValueError('primary key is required for delete_item')
            
            exec_namespaces = globals().copy()
            exec_namespaces["pk_validator"] = primary_key_field.converter
            
            fields = delete_item_options.get('fields')
            for field in fields:
                delete_item_handler_name = f'delete_{field.name}'
                primary_key_name = f"{manager_name}_id"
                func_sig = f"async def {delete_item_handler_name}(self, body: typing.Sequence, \n"
                func_sig += f"{primary_key_name}: typing.Annotated[typing.Any, pk_validator],\n"
                func_sig += "):\n"
                func_body = f"    return await self.delete_item(_id={primary_key_name}, \n"
                func_body += f"        field=self.scheme_cls.{field.in_scheme_name},\n"
                func_body += "        values=body\n"
                func_body += "    )\n"
                func_str = func_sig + func_body
                
                exec(func_str, exec_namespaces, result_namespaces)
                setattr(
                    manager_cls, delete_item_handler_name, 
                    result_namespaces[delete_item_handler_name]
                )

                if register:
                    register(
                        Method.DELETE,
                        '/{' + primary_key_name + '}/' + field.name,
                        getattr(manager_cls, delete_item_handler_name),
                    )

        return manager_cls

    return wrapper




import typing

from blue_firmament.dal.query_components import DALQueryComponent
from blue_firmament.dal.types import FieldLikeType
from blue_firmament.dal.utils import dump_field_like


class DALModifier(DALQueryComponent):
    ...


class LimitModifier(DALModifier):
    """结果只保留X个记录

    如果要限制更新、删除的记录数量，请在 update, delete 方法中使用 `limit` 参数。
    """

    def __init__(self, size: int = 1) -> None:
        super().__init__()
        self.__size = size

    def dump_to_sql(self) -> str:
        raise NotImplementedError('SingleFilter does not support SQL dump')

    def dump_to_postgrest(self) -> typing.Tuple[str, typing.Tuple[int]]:
        return ("limit", (self.__size,))

    def dump_to_postgrest_str(self) -> str:
        return f"limit.{self.__size}"


class RangeModifier(DALModifier):
    """结果只保留第X到Y个记录
    """

    def __init__(self, start: int = 0, end: int = 1) -> None:

        super().__init__()
        self.__start = start
        self.__end = end

    def dump_to_sql(self) -> str:
        raise NotImplementedError('SingleFilter does not support SQL dump')

    def dump_to_postgrest(self) -> typing.Tuple[str, typing.Tuple[int, int]]:
        return ("range", (self.__start, self.__end))

    def dump_to_postgrest_str(self) -> str:
        return f"range.{self.__start}.{self.__end}"


class OrderModifier(DALModifier):
    """按X字段排序结果
    """

    def __init__(self, field: FieldLikeType, *, desc: bool = False) -> None:
        """
        :param field: 排序字段
        :param desc: 是否降序排序（否则升序）
        """
        self._field = field
        self.__desc = desc

    def dump_to_sql(self) -> str:
        raise NotImplementedError('SingleFilter does not support SQL dump')

    def dump_to_postgrest(self) -> typing.Tuple[str, dict]:
        return ("order", {'column': dump_field_like(self._field), 'desc': self.__desc})

    def dump_to_postgrest_str(self) -> str:
        order_type = 'desc' if self.__desc else 'asc'
        return f"{dump_field_like(self._field)}.order.{order_type}"



"""Decorators for logging
"""

__all__ = [
    "log_manager_handler",
]

import copy
import inspect
import functools
import typing
from typing import Optional as Opt, Annotated as Anno, Literal as Lit
from .._types import AnnotatedDirective as AD
from ..utils.typing_ import is_annotated, get_annotated_args
from ..utils.inspect_ import args_to_kwargs_by_sig

P = typing.ParamSpec("P")
R = typing.TypeVar("R")


def log_manager_handler(func: typing.Callable[P, R]) -> typing.Callable[P, R]:
    
    """Decorator for logging manager handler

    Features
    --------
    - Log enterance and parameters
    - Log exitance and return value
    - Use handler level logger to replace manager level logger

    Disabled Parameters
    ^^^^^^^^^^^^^^^^^^^
    Disabled parameters will not be logged.

    Use ``typing.Annotated[type, AnnotatedDirective.NOLOG]`` to
    mark the parameter as disabled.
    """
    from ..manager.base import BaseManager

    func_name = func.__qualname__
    disabled_params = []

    # TODO get parameters annotated directive with xx
    param_sig = inspect.signature(func).parameters
    for param_name, param in param_sig.items():
        anno = param.annotation
        if is_annotated(anno):
            if get_annotated_args(anno)[0] == AD.NOLOG:
                disabled_params.append(param_name)

    @functools.wraps(func)
    def wrapper(
        *args: P.args, **kwargs: P.kwargs,
    ) -> R:
        
        self = args[0]
        if not isinstance(self, BaseManager):
            raise TypeError("First argument must be a BaseManager instance")

        to_log_parameters = kwargs.copy()
        if args:
            to_log_parameters.update(
                args_to_kwargs_by_sig(func, *args, offset=1)
            )
        if disabled_params:
            for param in disabled_params:
                to_log_parameters.pop(param, None)

        # bind logger
        logger = self._logger.bind(
            handler_name=func_name,
        )
        self_copy = copy.copy(self)
        self_copy._logger = logger

        # enter log
        logger.info("Enter manager handler", paramters=to_log_parameters)
        
        # call
        new_args = list(args) 
        new_args[0] = self_copy
        result = func(*new_args, **kwargs)  # type: ignore

        # exit log
        # FIXME not real exit for async function
        logger.info("Exit manager handler", result=result)

        return result

    return wrapper



'''BlueFirmament Manager Sub Package

Manager is a set of handlers.
'''

__all__ = [
    'BaseManager',
    'CommonManager',
    'PresetHandlerConfig'
]

from .base import BaseManager
from .common import CommonManager, PresetHandlerConfig



"""BlueFirmament TopLevel Exceptions

TODO exception factory to manage logger automatically
"""

__all__ = [
    'BlueFirmamentException',
    'BFExceptionTV',
    'InternalError',
    'ClientError',
    'RequestFailed',
    'ExternalError',
    'Retryable',
    'MaxRetriesExceeded',
    'ParamsInvalid',
    'AtLeastOne',
    'NotImplemented',
    'NotFound',
    'TaskHandlerNotFound',
    'DuplicateOrConflict',
    'Duplicate',
    'Conflict',
    'InvalidStatusTransition', 
    'Unauthorized',
    'Forbidden'
]

import enum
import abc
import typing
from typing import Optional as Opt, Annotated as Anno, Literal as Lit
from .utils.typing_ import JsonDumpable
from .task import TaskStatus, TaskID
from .log.main import get_logger
LOGGER = get_logger(__name__)
"""默认日志记录器
"""

if typing.TYPE_CHECKING:
    from .scheme import BaseScheme


BFExceptionTV = typing.TypeVar("BFExceptionTV", bound="BlueFirmamentException")
"""碧霄异常类型变量"""
class BlueFirmamentException(Exception, abc.ABC):

    """碧霄异常基类

    :ivar errmsg: 错误描述
    :ivar 

    """

    def __init__(self, 
        errmsg: str | dict = 'Exception occured', 
        *args, **kwargs
    ):
        if isinstance(errmsg, str):
            errmsg = {
                "errmsg": errmsg,
                **kwargs
            }
        
        self.errmsg: dict = errmsg

        super().__init__(errmsg, *args)

    def apply_to_task_result(self, response) -> None:
        """将异常序列化为响应体

        直接修改已有的响应对象的数据

        :param response: 响应对象
        """

    def dump_to_jsonable(self) -> JsonDumpable:
        """序列化为可JSON序列化的类型
        """
        return self.errmsg
    
    @property
    @abc.abstractmethod
    def task_status(self) -> TaskStatus:
        ...

    @classmethod
    def from_python_exception(cls, exception: Exception) -> "BlueFirmamentException":
        """从Python异常创建碧霄异常

        :param exception: Python异常
        """
        if isinstance(exception, ValueError):
            return ParamsInvalid()
        
        return InternalError()
    
    @property
    def name(self) -> str:
        """异常名称"""
        return self.__class__.__name__
    
    def dump_details_to_str(self) -> str:
        """将详细信息序列化为字符串
        """
        return '\n'.join([
            f"{k}: {v}"
            for k, v in self.errmsg.items()
        ])
    
    def dump_details_to_dict(self) -> dict:
        """将详细信息序列化为字典
        """
        return self.errmsg
    
    def __str__(self) -> str:
        return f"{self.name}: \n{self.dump_details_to_str()}"


class InternalError(BlueFirmamentException):
    """Exception raised cause of server's fault.
    """

    @property
    def task_status(self) -> TaskStatus:
        return TaskStatus.INTERNAL_SERVER_ERROR


class ClientError(BlueFirmamentException):
    """Exception raised cause of client's fault
    """

    @property
    def task_status(self) -> TaskStatus:
        return TaskStatus.BAD_REQUEST


class RequestFailed(InternalError):

    """请求失败

    请求没有到达目标服务
    """

    def __init__(self, 
        response,
        *args, **kwargs
    ):

        super().__init__(
            {
                "response": response
            }, 
            logger, *args, **kwargs
        )


class ExternalError(BlueFirmamentException):

    """外部错误

    任何外部应用程序与服务导致的异常都将转换为该异常
    """

    def __init__(self, 
        service_name: str, 
        logger=LOGGER,
        **errinfo: typing.Any,
    ):
        
        """
        :param service_name: 外部服务名称
        :param errcode: 错误码
            
            可以是错误码、错误信息等
        """

        super().__init__({
            "service_name": service_name,
            **errinfo
        }, logger)

    def __getitem__(self, key: str) -> typing.Any:
        """获取错误信息"""
        return self.errmsg[key]

    @property
    def task_status(self):
        return TaskStatus.SERVICE_UNAVAILABLE


class Retryable(BlueFirmamentException):
    """The exception can be eliminated by retrying the operation.
    """

    def __init__(
        self,
        delay: Opt[float] = None
    ):
        """
        :param delay: retry after this many seconds
        """
        super().__init__("please retry after %s seconds" % delay)
        self.__delay = delay

    @property
    def delay(self) -> Opt[float]:
        return self.__delay

    @property
    def task_status(self):
        return TaskStatus.SERVICE_UNAVAILABLE


class MaxRetriesExceeded(BlueFirmamentException):

    def __init__(self):
        super().__init__("max retries exceeded")

    @property
    def task_status(self) -> TaskStatus:
        return TaskStatus.SERVICE_UNAVAILABLE


class ParamsInvalid(ClientError, ValueError):

    """参数无效

    场景
    ------
    - 类型不合法
    - 值不合法
    - 缺少参数
    - 通过参数获得（计算、数据库访问）的值不合法
    """

    def __init__(
        self,
        msg: str = '',
        **params
    ):
        super().__init__({
            "message": "Invalid parameter(s), %s" % msg,
            "parameters": params
        })

    @property
    def task_status(self):
        return TaskStatus.UNPROCESSABLE_ENTITY
    

class ParamRequired(ClientError, ValueError):
    """Exception for not giving required parameter.
    """

    def __init__(self, param_name: str, **kwargs):
        super().__init__("parameter required", param=param_name)


class AtLeastOne(ParamsInvalid):

    """至少需要一个参数

    在这些参数中，至少提供一个（有效）
    """

    def __init__(self, 
        *params: typing.Any,
        logger=LOGGER,
    ):
        
        """
        :param params: 参数名称
        """
        self.__params = params

    def __str__(self) -> str:
        
        return f"{self.name}: at least one of following parameters should be provided:\n" + \
            "\n    ".join(self.__params)
    

class NotImplemented(InternalError, NotImplementedError):

    def __init__(self, 
        service_name: str, 
        *args, **kwargs
    ):
        
        """
        :param service_name: 未实现的服务名
        """
        errmsg = "service %s is not implemented" % service_name
        super().__init__(errmsg, logger, *args, **kwargs)

    @property
    def task_status(self) -> TaskStatus:
        return TaskStatus.NOT_IMPLEMENTED

class NotFound(ClientError, KeyError):
    """未找到
    """

    @property
    def task_status(self) -> TaskStatus:
        return TaskStatus.NOT_FOUND

class TaskHandlerNotFound(NotFound):
    """没有该任务的处理器
    """

    def __init__(self, task_id: TaskID):
        super().__init__("task handler not found", task_id=task_id)

    @property
    def task_id(self):
        return self.errmsg.get("task_id", None)

class DuplicateOrConflict(InternalError):

    """冲突或重复
    """

    def __init__(self, 
        *resource: tuple[typing.Any, typing.Any],
        operation: str,
        msg: str = '',
    ):
        
        """
        :param resource: 资源的的元组
            用（资源名，资源标识符）表示一个资源实例
        :param operation: 正在对资源进行的操作
        :param errmsg: 附加描述
        """

        super().__init__(
            f"Cannot perform {operation} on {resource[0]}: {resource[1]} cause: \n{msg}"
        )

    @classmethod
    def from_scheme(cls, 
        scheme: "BaseScheme",
        operation: str,
        errmsg: str = '',
    ) -> 'DuplicateOrConflict':
        """从数据模型 实例创建冲突异常
        """
        return cls(
            (scheme.dal_path(), scheme[scheme._get_key_field()]),
            operation=operation, msg=errmsg,
        )

    @property
    def task_status(self) -> TaskStatus:
        return TaskStatus.CONFLICT


class Duplicate(DuplicateOrConflict):

    def __init__(self, msg: str):
        self.__msg = msg

    def dump_details_to_str(self) -> str:
        return self.__msg


class Conflict(DuplicateOrConflict):

    def __init__(self, msg: str, **kwargs):
        self.__msg = msg
        self.__context = kwargs

    def dump_details_to_str(self) -> str:
        return self.__msg


class InvalidStatusTransition(DuplicateOrConflict):

    """无效的状态转换

    不可以从当前状态切换到目标状态
    """
    def __init__(self,
        status_name: str,
        current_status, target_status
    ):
        
        self.__status_name = status_name
        self.__current_status = current_status
        self.__target_status = target_status

    @classmethod
    def from_enum_member(cls,
        current: enum.Enum, target: enum.Enum
    ) -> typing.Self:
        
        """从枚举成员创建
        """
        return cls(
            current.__class__.__name__,
            current.value, target.value
        )

    def dump_details_to_str(self) -> str:
        return f"cannot transist {self.__status_name} from {self.__current_status} to {self.__target_status}"


class Unauthorized(InternalError):
    """未认证
    """

    def __init__(self, 
        msg: Opt[str] = None,
        identity: Opt[str] = None,
        *args, **kwargs
    ):
        """
        :param identity: 导致未认证时使用的凭证（如果有）
        """
        super().__init__(
            'you are using identity (token): \n%s' % identity
        )

    @property
    def task_status(self) -> TaskStatus: 
        return TaskStatus.UNAUTHORIZED

        
class Forbidden(InternalError):

    """无权操作
    """

    def __init__(self, 
        msg: str,
    ):
        super().__init__(msg)

    @property
    def task_status(self): 
        return TaskStatus.FORBIDDEN


# class UnavailableForLegalReasons(BlueFirmamentException):

#     def __init__(self, errmsg, *args, **kwargs) -> None:

#         super().__init__(451, "Unavailable For Legal Reasons, %s" % errmsg, *args, **kwargs)


# class TooManyRequests(BlueFirmamentException):

#     def __init__(self, errmsg, *args, **kwargs) -> None:

#         super().__init__(429, "Too many request, limit reached. %s" % errmsg, *args, **kwargs)



'''Utils related to type checking and type hinting.'''

import typing
import types
import inspect
import datetime

if typing.TYPE_CHECKING:
    from ..task.result import JsonBody

if typing.TYPE_CHECKING:
    from ..scheme import BaseScheme

def is_annotated(tp: typing.Type) -> typing.TypeGuard[typing.Annotated]:

    '''Check if the tp is typing.Annotated

    Example
    -------
    >>> is_annotated(typing.get_origin(typing.Annotated[str, ...]))
    True
    >>> is_annotated(typing.Annotated[str, ...])
    True
    '''
    if tp is not typing.Annotated:
        return typing.get_origin(tp) is typing.Annotated
    else:
        return True
    
def get_annotated_args(tp: typing.Annotated) -> typing.Tuple[typing.Any, ...]:

    '''获取 Annotated 的参数

    Example
    -------
    >>> get_annotated_args(typing.Annotated[str, ...])
    ()
    >>> get_annotated_args(typing.Annotated[str, 1, 2])
    (1, 2)
    '''
    return typing.get_args(tp)[1:]
    
    return ()


@typing.overload
def get_origin(tp: types.UnionType) -> types.UnionType:
    ...
@typing.overload
def get_origin(tp: typing.Type) -> typing.Type:
    ...
def get_origin(tp: typing.Type | types.UnionType) -> typing.Type | types.UnionType:

    '''
    获取复杂类型的原始类型

    Behavior
    --------
    - typing.Annotated -> args[0]
    - typing.NewType -> __supertype__
    - typing.Union, types.UnionType 不会被处理，直接返回

    Example
    -------
    >>> get_origin(typing.Annotated[str, ...])
    str
    >>> get_origin(typing.NewType('MyStr', str))
    str
    >>> get_origin(typing.NewType('MyStr', typing.NewType('MyInt', int)))
    int
    >>> get_origin(typing.Union[str, int])
    typing.Union[str, int]
    >>> get_origin(str | int)
    str | int
    >>> get_origin(str)
    str
    '''
    if isinstance(tp, typing.NewType):
        tp_ = tp.__supertype__
        while isinstance(tp_, typing.NewType):
            tp_ = tp_.__supertype__
        return tp_
    
    res = typing.get_origin(tp) or tp

    if res is typing.Union or res is types.UnionType:
        return tp

    if res is typing.Annotated:
        return typing.get_args(tp)[0]
    
    return res


T = typing.TypeVar('T', bound=typing.Type)
def ismethodorigin(func, origin: T) -> typing.TypeGuard[T]:
    
    '''判断是否为某个类的方法

    包括实例方法、类方法：
    
    ```python
    class A:
        def method(self):
            pass
        @classmethod
        def cls_method(cls):
            pass
    
    A.method # no
    A.cls_method # yes
    a = A()
    a.method # yes
    ```

    :param param: 要检查的参数
    :param origin: 要检查的类
    '''

    if inspect.ismethod(func):
        return isinstance(func.__self__, origin)
    elif isclassmethod(func):
         return func.__self__ is origin
    
    return False

def isclassmethod(func: typing.Any) -> bool:

    '''Check if the function is a classmethod
    '''
    return inspect.ismethod(func) and hasattr(func, '__self__') and func.__self__ is not None


type JsonDumpable = typing.Union[
    str, int, float, bool, None, 
    datetime.datetime,
    typing.List['JsonDumpable'], 
    typing.Tuple['JsonDumpable', ...],
    typing.Dict[str, 'JsonDumpable'],
    typing.Set['JsonDumpable'],
    "BaseScheme",
    "JsonBody",
]
'''可以序列化为JSON的类型

其中BaseScheme实际上不能被json.dumps处理， 
需要通过我们自定义的json_dumps来处理，调用 :func:`utils.json.dumps_to_json`
'''
def is_json_dumpable(val: typing.Any) -> typing.TypeGuard[JsonDumpable]:

    """

    TODO
    ----
    - 对于容器类型，没有进一步对元素判断
    """
    if val is None:
        return True

    from ..scheme import BaseScheme
    from ..task.result import JsonBody
    if isinstance(val, (
        str, int, float, bool,
        list, tuple, dict, set,
        datetime.datetime,
        BaseScheme, JsonBody
    )):
        return True
    
    return False

def is_union_type(tp):
    """Checks if a type hint is a union type (typing.Union or |)."""
    origin = typing.get_origin(tp)
    return origin is typing.Union or origin is types.UnionType


def add_type_to_namespace(type_: typing.Type, namespace: dict):
    '''将类型添加到命名空间

    :param type: 要添加的类型
    :param namespace: 命名空间
    '''
    try:
        namespace[type_.__name__] = type_
    except AttributeError:
        raise ValueError('Cannnot process this type %s' % type_)


def is_mutable(value: typing.Any) -> bool:
    '''判断一个值是否为可变对象

    :param value: 要判断的值
    '''
    if isinstance(value, (list, dict, set)):
        return True
    elif hasattr(value, '__hash__'):
        return False  # not strict enough
    elif hasattr(value, '__dict__'):
        return True  # need verification
    else:
        return False


# TODO rename to `try_issubclass`
def safe_issubclass(
    obj: typing.Any, classinfo: T
) -> typing.TypeGuard[T]:
    return isinstance(obj, type) and issubclass(obj, classinfo)


def is_namedtuple(obj: typing.Type) -> typing.TypeGuard[typing.NamedTuple]:
    return safe_issubclass(obj, tuple) and hasattr(obj, '_fields')




import typing
from typing import Annotated as Anno, Optional as Opt, Literal as Lit
import redis.asyncio as redis
from .base import KVLikeDataAccessLayer, QueueLikeDataAccessLayer, PubSubLikeDataAccessLayer, PubSubMessage
from .types import StrictDALPath
from ..data.settings.dal import get_setting as get_dal_setting
from ..exceptions import NotFound


class RedisDAL(
    KVLikeDataAccessLayer,
    QueueLikeDataAccessLayer,
    PubSubLikeDataAccessLayer
):

    def __init_subclass__(
        cls,
        host: str = "localhost",
        port: int = 6379,
        password: Opt[str] = None,
        db: int = 0,
        **kwargs
    ):
        cls.__host = host
        cls.__port = port
        cls.__password = password

        return super().__init_subclass__(
            default_path=StrictDALPath((db,)),
            **kwargs
        )

    def __init__(self, **kwargs) -> None:
        self._client = redis.Redis(
            host=self.__host,
            port=self.__port,
            password=self.__password,
            db=self.default_path[0]
        )
        self._pubsub = self._client.pubsub(ignore_subscribe_messages=True)
        super().__init__(**kwargs)

    async def close(self):
        await self._pubsub.aclose()
        await self._client.aclose()

    def get(self, key: str) -> Opt[typing.Any]:
        return self._client.get(key)

    async def set(self, key: str, value: typing.Any) -> None:
        await self._client.set(key, value)

    async def push(self, item: bytes, queue_name: Opt[str] = None) -> None:
        await self._client.lpush(queue_name or self.__queue_name__, item)

    async def pop(
        self,
        queue_name: Opt[str] = None,
        wait: bool = True
    ) -> bytes:
        queue_name = queue_name or self.__queue_name__

        if not wait:
            res = await self._client.lpop(queue_name)
        else:
            _, res = typing.cast(
                tuple[bytes, bytes],
                await self._client.blpop([queue_name], timeout=0)
            )

        if res is None:
            raise NotFound(f"No available items in queue {queue_name}")
        if isinstance(res, str):
            return res.encode('utf-8')

        return res

    async def publish(self, item: bytes, *channel_names: str) -> None:
        if not channel_names:
            channel_names = (self.__channel_name__,)
        for channel_name in channel_names:
            await self._client.publish(channel_name, item)

    async def subscribe(self, *channel_names: str) -> None:
        if not channel_names:
            channel_names = (self.__channel_name__,)
        await self._pubsub.subscribe(*channel_names)

    async def unsubscribe(self, *channel_names: str) -> None:
        if not channel_names:
            channel_names = (self.__channel_name__,)
        await self._pubsub.unsubscribe(*channel_names)
        await self._pubsub.close()

    async def get_message(self, timeout: Opt[int] = None) -> PubSubMessage:
        message = None
        while message is None:
            message = await self._pubsub.get_message(timeout=timeout)

        return PubSubMessage(
            channel=message["channel"],
            data=message["data"],
        )

    async def listen(self) -> typing.AsyncIterable[PubSubMessage]:
        async for message in self._pubsub.listen():
            if message['type'] == "message":
                yield PubSubMessage(
                    channel=message['channel'],
                    data=message['data']
                )




from ...setting import Setting, make_setting_singleton


class DALSetting(Setting):

    _setting_name: str = "dal"

    postgrest_url: str = 'http://localhost:3000'
    '''默认PostgREST的URL地址'''
    postgrest_default_schema: str = 'public'
    '''默认PostgREST的默认表组（schema）'''
    postgres_default_table: str = 'default_table'
    '''默认PostgreSQL的默认表名（table）'''
    postgrest_anonymous_token: str = ''
    '''默认PostgREST的匿名访问令牌（仅在Supabase Postgrest中使用）'''
    postgrest_service_token: str = ''
    '''默认PostgREST的服务访问令牌（仅在Supabase Postgrest中使用）'''

    redis_host: str = 'localhost'
    '''默认Redis服务器的主机名'''
    redis_port: int = 6379
    '''默认Redis服务器的端口号'''
    redis_password: str = ''
    '''默认Redis服务器的密码'''
    redis_db: int = 0
    '''默认Redis服务器的数据库编号'''


get_setting, set_setting = make_setting_singleton(DALSetting())




"""This module contains the operators for data access layer.
"""

import abc
import typing
from typing import Optional as Opt

from . import DALQueryComponent


class DALOperator(DALQueryComponent):
    """Abstract base class for DAL operators.

    Operators are logical constructs that combine or modify filters.
    """

class NotOperator(DALOperator):

    def dump_to_sql(self) -> str:
        return "NOT"

    def dump_to_postgrest(self) -> typing.Tuple[str, tuple]:
        return ("not_", ())

    def dump_to_postgrest_str(self) -> str:
        return "not"


class OrOperator(DALOperator):

    def __init__(self, *filters) -> None:
        super().__init__()
        if len(filters) < 2:
            raise ValueError("OrOperator requires at least two filters")
        self.__filters = filters

    def dump_to_sql(self) -> str:
        return "OR"

    def dump_to_postgrest(self) -> tuple[str, tuple]:
        return ("or_", (",".join(
            filter_.dump_to_postgrest_str()
            for filter_ in self.__filters
        ),))

    def dump_to_postgrest_str(self) -> str:
        return f"or({','.join(
            filter_.dump_to_postgrest_str() 
            for filter_ in self.__filters
        )})"



import enum

from blue_firmament.task import TaskStatus


class MIMEType(enum.Enum):
    JSON = 'application/json'
    XML = 'application/xml'
    FORM = 'application/x-www-form-urlencoded'
    TEXT = 'text/plain'
    EVENT_STREAM = 'text/event-stream'
    BINARY = 'application/octet-stream'


class HTTPHeader(enum.Enum):
    AUTHORIZATION = "authorization"
    CONTENT_TYPE = "content-type"
    CONTENT_ENCODING = "content-encoding"
    ACCEPT = "accept"
    ACCEPT_CHARSET = "accept-charset"
    COOKIE = "cookie"
    SET_COOKIE = "set-cookie"
    CACHE_CONTROL = "cache-control"
    CONNECTION = "connection"
    TRACE_ID = "x-trace-id"
    CLIENT_ID = "x-client-id"


TStatus2HCode: dict[TaskStatus, int] = {
    TaskStatus.OK: 200,
    TaskStatus.DELETED: 204,
    TaskStatus.CREATED: 201,
    TaskStatus.BAD_REQUEST: 400,
    TaskStatus.CONFLICT: 409,
    TaskStatus.FORBIDDEN: 401,
    TaskStatus.NOT_FOUND: 404,
    TaskStatus.UNAUTHORIZED: 403,
    TaskStatus.UNPROCESSABLE_ENTITY: 422,
    TaskStatus.UNAVAILABLE_FOR_LEGAL_REASONS: 451,
    TaskStatus.INTERNAL_SERVER_ERROR: 500,
}
"""Map Task status code to HTTP status code
"""



'''碧霄日志模块

See :doc:`docs/log`
'''

import logging
import typing
import structlog
from typing import Optional as Opt, Annotated as Anno
from ..data.settings.log import get_setting as get_log_setting

if typing.TYPE_CHECKING:
    from . import LoggerT


logging.basicConfig(level=get_log_setting().log_level)

logger_factory = None
log_file = get_log_setting().log_file
if log_file is not None:
    logger_factory = structlog.WriteLoggerFactory(
        file=open(log_file, mode="wt"),
    )
else:
    logger_factory = structlog.stdlib.LoggerFactory()

structlog.configure(
    processors=[
        structlog.stdlib.filter_by_level,
        structlog.processors.TimeStamper(fmt="iso", key="datetime"),
        structlog.processors.dict_tracebacks,
        structlog.processors.CallsiteParameterAdder(
            parameters=[
                structlog.processors.CallsiteParameter.FILENAME,
                structlog.processors.CallsiteParameter.LINENO,
            ]
        ),
        structlog.processors.JSONRenderer(),
    ],
    context_class=dict,
    logger_factory=logger_factory,
    wrapper_class=structlog.stdlib.BoundLogger,
    cache_logger_on_first_use=get_log_setting().cache_logger_on_first_use,
)



def get_logger(name) -> "LoggerT":
    return structlog.get_logger(name)

def bind_logger_contextvars(**contextvars) -> None:
    structlog.contextvars.bind_contextvars(**contextvars)

def clear_logger_contextvars() -> None:
    structlog.contextvars.clear_contextvars()




"""PostgrestDAL
"""

import postgrest

from ..task.context import ExtendedTaskContext
from .._types import _undefined
from ..scheme.converter import SchemeConverter
from .utils import dump_query_coms_like
from ..exceptions import Unauthorized
from ..utils.typing_ import safe_issubclass
from ..exceptions import NotFound
from ..scheme.field import Field, FieldValueProxy, FieldValueTV
from blue_firmament.dal.query_components.filters import *
from .base import TableLikeDataAccessLayer, DataAccessLayerWithAuth
from .. import __version__, __name__ as __package_name__
from .types import (
    DALPath, FieldLikeType, QueryComLikeType, StrictDALPath
)
from .query_components import DALQueryComponent
from ..utils.main import call_as_sync
from ..utils.enum_ import dump_enum
from ..scheme import BaseScheme, SchemeTV


class PostgrestDAL(TableLikeDataAccessLayer, DataAccessLayerWithAuth):
    """Access data through PostgREST API.

    Examples
    --------
    .. code-block:: python
        from blue_firmament.dal.postgrest import PostgrestDAO

        class SupaAnonPostgrestDAO(PostgrestDAO,
            url=get_supabase_setting().supabase_url,
            apikey=get_supabase_setting().anon_key,
            default_table="profile"
        ):
            pass
    """
    
    def __init_subclass__(
        cls,
        url: str,
        apikey: str,
        default_table: str | enum.Enum,
        default_schema: str | enum.Enum = "public",
        **kwargs
    ) -> None:
        """
        :param apikey: Supabase API Key.
            Use anon key for authenticated or anonymous user,
            use serv key for service_role user.
        """
        cls.__url = url
        cls.__apikey = apikey

        return super().__init_subclass__(
            default_path=StrictDALPath((dump_enum(default_table), dump_enum(default_schema))),
            **kwargs
        )

    def __post_init__(self):
        self._client = postgrest.AsyncPostgrestClient(
            base_url=self.__url,
            schema=dump_enum(self.default_path[1]),
            headers={
                'X-Client-Info': f'{__package_name__}/{__version__}',
                'apiKey': self.__apikey,
                **({"authorization": f'Bearer {self._auth_session.access_token}'}
                    if self._auth_session else {})
            }
        )

    async def close(self) -> None:
        await self._client.aclose

    def set_schema(self, schema: str) -> None:
        """设置操作的表组（schema）"""
        self._client.schema(schema)

    def __get_base_query_from_path(self, path: DALPath | None = None):
        """从路径中获取查询对象"""
        dp = self.dump_path(path)
        return self._client.schema(dp[1]).from_table(dp[0])
    

    QueryTV = typing.TypeVar('QueryTV', 
        postgrest.AsyncQueryRequestBuilder,
        postgrest.AsyncFilterRequestBuilder,
        postgrest.AsyncSelectRequestBuilder
    )
    def __apply_filters_to_base_query(
        self,
        base_query: QueryTV,
        query_coms: typing.Iterable[DALQueryComponent]
    ) -> QueryTV:
        """将过滤器应用到查询对象

        :param base_query: 基础查询对象
        :param query_coms: 过滤器列表
        """
        for query_com in query_coms:
            dumped_query_com = query_com.dump_to_postgrest()
            req_builer_method = getattr(base_query, dumped_query_com[0])
            if isinstance(dumped_query_com[1], tuple):
                base_query = req_builer_method(*dumped_query_com[1])
            elif isinstance(dumped_query_com[1], dict):
                base_query = req_builer_method(**dumped_query_com[1])
            else:
                base_query = req_builer_method()
        return base_query

    async def __execute_query(self, query: QueryTV):
        """执行编辑好的请求

        处理这些异常：
        - PGRST301 -> Unauthorized
        - code42501 -> New row violates row-level security policy
        """
        try:
            return await query.execute()
        except postgrest.APIError as e:
            if e.code == 'PGRST301':  # JWT expired
                raise Unauthorized("token expired", self._auth_session.access_token)
            
            raise e

    @typing.overload
    async def insert(self,
        to_insert: dict,
        path: typing.Optional[DALPath] = None,
        exclude_key: bool = True,
    ) -> dict:
        ...
    @typing.overload
    async def insert(self,
        to_insert: SchemeTV,
        path: typing.Optional[DALPath] = None,
        exclude_key: bool = True,
    ) -> SchemeTV:
        ...
    async def insert(
        self,
        to_insert: dict | SchemeTV,
        path = None,
        exclude_natural_key: bool = True,
    ) -> dict | SchemeTV:
        if isinstance(to_insert, BaseScheme) and path is None:
            path = to_insert.dal_path()
        
        processed_to_insert: dict
        if isinstance(to_insert, BaseScheme):
            processed_to_insert = to_insert.dump_to_dict(
                exclude_natural_key=exclude_natural_key
            )
        else:
            processed_to_insert = to_insert
        
        base_query = self.__get_base_query_from_path(path)
        query = base_query.insert(
            json=processed_to_insert
        )
        res = await self.__execute_query(query)
        
        if isinstance(to_insert, BaseScheme):
            sc = SchemeConverter(scheme_cls=to_insert.__class__)
            return sc(
                res.data[0],
                **to_insert.dump_to_dict(only_private=True)
            )
        elif isinstance(to_insert, dict):
            return res.data[0]
        
        assert False

    FieldValueTV = typing.TypeVar('FieldValueTV')

    @typing.overload
    async def select(
        self,
        to_select: typing.Type[SchemeTV],
        *filters: QueryComLikeType,
        path: typing.Optional[DALPath] = None,
        task_context: Opt[ExtendedTaskContext] = None,
    ) -> typing.Tuple[SchemeTV, ...]:
        ...
    @typing.overload
    async def select(
        self,
        to_select: "Field[FieldValueTV]",
        *filters: QueryComLikeType,
        path: typing.Optional[DALPath] = None,
        task_context: Opt[ExtendedTaskContext] = None,
    ) -> typing.Tuple[FieldValueTV, ...]:
        ...
    @typing.overload
    async def select(
        self,
        to_select: typing.Iterable[FieldLikeType] | None,
        *filters: QueryComLikeType,
        path: typing.Optional[DALPath] = None,
        task_context: Opt[ExtendedTaskContext] = None,
    ) -> typing.Tuple[dict, ...]:
        ...
    async def select(
        self,
        to_select: typing.Union[
            typing.Type[SchemeTV], 
            "Field[FieldValueTV]",
            typing.Iterable[FieldLikeType],
            None
        ],
        *query_coms: QueryComLikeType,
        path: typing.Optional[DALPath] = None,
        task_context: Opt[ExtendedTaskContext] = None,
    ) -> typing.Union[
        typing.Tuple[SchemeTV, ...],
        typing.Tuple[FieldValueTV, ...],
        typing.Tuple[dict, ...]
    ]:
        # process to_select to fields
        if to_select is None:
            fields = ("*",)
        elif isinstance(to_select, Field):
            fields = (to_select.name,)
        elif isinstance(to_select, typing.Iterable):
            fields = tuple(
                dump_field_like(i) for i in to_select
            )
        elif issubclass(to_select, BaseScheme):
            fields = ("*",)
        else:
            raise ValueError(f"Invalid type for to_select, {type(to_select)}")

        # preprocess path
        if path is None:
            if isinstance(to_select, type) and safe_issubclass(to_select, BaseScheme):
                path = to_select.dal_path()
            elif isinstance(to_select, Field):
                path = to_select.scheme_cls.dal_path()

        # process query components
        prcesd_query_coms: typing.Iterable[DALQueryComponent] = dump_query_coms_like(
            *query_coms, scheme_like=to_select
        )
        ref_fields = set()
        for i, query_com in enumerate(prcesd_query_coms):
            if isinstance(query_com, DALFilter):
                filter_dal_path = query_com.dal_path
                if (
                    filter_dal_path is not None and
                    filter_dal_path[0] != path[0]
                ):
                    ref_fields.add(f"{filter_dal_path[0]}({query_com.field_name})")
                    prcesd_query_coms[i] = query_com.fork(use_fqfn=True)

        # construct query
        base_query = self.__get_base_query_from_path(path)
        base_query = base_query.select(*fields, *ref_fields)
        query = self.__apply_filters_to_base_query(base_query, prcesd_query_coms)
        res = await self.__execute_query(query)

        if len(res.data) == 0:
            raise NotFound(str(path))  # TODO self, filters=processed_filters
        
        # parse res to the same as to_selec
        if isinstance(to_select, Field):
            return tuple(
                to_select.load_val(i[to_select.name])
                for i in res.data
            )
        elif safe_issubclass(to_select, BaseScheme): 
            sc = SchemeConverter(scheme_cls=to_select)
            return tuple(
                sc(
                    instance_dict, 
                    _task_context=task_context or _undefined
                )
                for instance_dict in res.data
            )  # type: ignore
        else:
            return tuple(*res.data)
    
    async def delete(
        self,
        to_delete: SchemeTV | typing.Type[SchemeTV],
        *query_coms: QueryComLikeType,
        path: Opt[DALPath] = None,
     ) -> None:
        
        if path is None:
            if isinstance(to_delete, BaseScheme) or issubclass(to_delete, BaseScheme):
                path = to_delete.dal_path()
        
        if not query_coms:
            if isinstance(to_delete, BaseScheme):
                query_coms += (to_delete.key_eqf,)
        
        base_query = self.__get_base_query_from_path(path)
        base_query = base_query.delete()
        query = self.__apply_filters_to_base_query(
            base_query, dump_query_coms_like(*query_coms, scheme_like=to_delete)
        )
        res = await self.__execute_query(query)

        if len(res.data) == 0:
            raise DeleteFailure(path, self)  # TODO add NoEffect (checkout filters, RLS)

    @typing.overload
    async def update(
        self,
        to_update: SchemeTV,
        *filters: DALQueryComponent,
        path: Opt[DALPath] = None,
        only_dirty: bool = True,
        exclude_natural_key: bool = True,
    ) -> SchemeTV:
        ...
    @typing.overload
    async def update(
        self,
        to_update: dict,
        *filters: DALQueryComponent,
        path: Opt[DALPath] = None,
        only_dirty: bool = True,
        exclude_natural_key: bool = True,
    ) -> dict:
        ...
    @typing.overload
    async def update(
        self,
        to_update: "FieldValueProxy[FieldValueTV]" | FieldValueTV,
        *filters: DALQueryComponent,
        path: Opt[DALPath] = None,
        only_dirty: bool = True,
        exclude_natural_key: bool = True,
    ) -> FieldValueTV:
        ...
    @typing.overload
    async def update(
        self,
        to_update: typing.Tuple[Field[FieldValueTV], FieldValueTV],
        *filters: DALQueryComponent,
        path: Opt[DALPath] = None,
        only_dirty: bool = True,
        exclude_natural_key: bool = True,
    ) -> FieldValueTV:
        ...
    async def update(
        self,
        to_update: typing.Union[
            dict, SchemeTV,
            "FieldValueProxy[FieldValueTV]",
            FieldValueTV,
            typing.Tuple[Field[FieldValueTV], FieldValueTV]
        ],
        *query_coms: DALQueryComponent,
        path: Opt[DALPath] = None,
        only_dirty: bool = True,
        exclude_natural_key: bool = True,
    ) -> typing.Union[
        dict,
        SchemeTV,
        FieldValueTV,
    ]:
        # preprocess path
        if path is None: 
            if isinstance(to_update, BaseScheme):
                path = to_update.dal_path()
            elif isinstance(to_update, FieldValueProxy):
                path = to_update.scheme.dal_path()

        # preprocess filters
        if not query_coms:
            if isinstance(to_update, BaseScheme):
                query_coms += (
                    to_update.key_eqf,
                )
            elif isinstance(to_update, FieldValueProxy):
                query_coms += (
                    to_update.scheme.key_eqf,
                )
        
        # process to_update
        processed_to_update: typing.Dict[str, typing.Any]
        if isinstance(to_update, BaseScheme):
            processed_to_update = to_update.dump_to_dict(
                only_dirty=only_dirty,
                exclude_natural_key=exclude_natural_key
            )
        elif isinstance(to_update, FieldValueProxy):
            processed_to_update = {to_update.field.name: to_update.obj}
        elif isinstance(to_update, tuple):
            if len(to_update) != 2:
                raise ValueError(f"Invalid tuple length for to_update, {len(to_update)}")
            processed_to_update = { to_update[0].name: to_update[1] }
        elif isinstance(to_update, dict):
            processed_to_update = {
                k.name if isinstance(k, Field) else k: v
                for k, v in to_update.items()
            }
        else:
            raise ValueError(f"Invalid type for to_update, {type(to_update)}")

        base_query = self.__get_base_query_from_path(path)
        base_query = base_query.update(json=processed_to_update)
        query = self.__apply_filters_to_base_query(base_query, query_coms)
        res = await self.__execute_query(query)

        if len(res.data) == 0:
            raise UpdateFailure(path, self)  # TODO add NoEffect (checkout filters, RLS)
        
        # parse res to the same as to_update
        if isinstance(to_update, BaseScheme):
            sc = SchemeConverter(scheme_cls=to_update.__class__)
            return sc(
                value=res.data[0],
                **to_update.dump_to_dict(only_private=True)
            )
        elif isinstance(to_update, FieldValueProxy):
            return to_update.field.load_val(res.data[0][to_update.field.name])
        elif isinstance(to_update, tuple):
            return res.data[0][to_update[0].name]
        elif isinstance(to_update, dict):
            return res.data[0]
        
        assert False

    async def upsert(self):  # TODO
        return await super().upsert()



import asyncio

from .._types import TaskRegistriesT
from ..task.context import CommonTaskContext
from ..task.context import ExtendedTaskContext, BaseTaskContext
from ..log.main import get_logger
from ..task.result import TaskResult
from ..transport.base import (
    BaseTransporter
)
from ..task import Task
from ..task.registry import TaskRegistry
from .middleware import BaseMiddleware, MiddlewaresT
from blue_firmament.dal.query_components.filters import *

if typing.TYPE_CHECKING:
    from structlog.stdlib import BoundLogger
    from ..manager import BaseManager


class BlueFirmamentApp:

    def __init__(
        self,
        name: str = "",
        registries: Opt[TaskRegistriesT] = None,
        transporters: Opt[typing.Iterable[BaseTransporter]] = None,
        middlewares: Opt[MiddlewaresT] = None,
        task_context_cls: type[ExtendedTaskContext] = CommonTaskContext,
    ):
        """
        :param registries:
            If None, will create an empty registry for each transporter.
        """
        self.__transporters: set[BaseTransporter] = set(transporters or ())
        self.__task_registries: TaskRegistriesT = registries or {
            transporter: TaskRegistry(name=str(transporter))
            for transporter in self.__transporters
        }
        self.__task_context_cls: type[ExtendedTaskContext] = task_context_cls
        self.__middlewares: MiddlewaresT = middlewares or []
        self.__logger = get_logger(f"BFApp[{name}]").bind(
            app_name=name
        )

    @property
    def _logger(self) -> "BoundLogger":
        return self.__logger

    def add_transporter(
        self,
        transporter: BaseTransporter,
        registry: Opt[TaskRegistry] = None
    ):
        """Add a transporter and its TaskRegistry.
        """
        self.__transporters.add(transporter)
        self.__task_registries[transporter] = registry or TaskRegistry(
            name=transporter.name
        )

    def add_manager(self, manager: type["BaseManager"]):
        """Merge the manager's task registries.
        """
        for transporter, task_entry in manager.__task_registries__.items():
            self.__task_registries[transporter].merge(task_entry)

    def add_managers(self, *managers: type["BaseManager"]):
        for manager in managers:
            self.add_manager(manager)

    def run(self):
        """Start the application.
        """
        event_loop = asyncio.new_event_loop()
        try:
            asyncio.set_event_loop(event_loop)

            for transport in self.__transporters:
                event_loop.create_task(transport.start())
            
            event_loop.run_forever()
        except KeyboardInterrupt:
            self._logger.info('Stop for KeyboardInterrupt')
        finally:
            event_loop.stop()
            event_loop.close()

    async def handle_task(
        self,
        transporter: BaseTransporter | str,
        task: Task,
        task_result: TaskResult
    ):
        task_entry = self.__task_registries[transporter].lookup(task.id)
        middlewares: MiddlewaresT = self.__middlewares + [task_entry]
        task_context = self.__task_context_cls(BaseTaskContext(
            task=task,
            task_result=task_result,
            base_logger=self._logger
        ))
        BaseTaskContext.set_contextvar(task_context)
        await BaseMiddleware.run_middlewares(middlewares, task_context)



"""DataAccessLayer

Design Doc: :doc:`/design/dal`
"""

__all__ = [
    "DALPath", "KeyableType",
    "DataAccessLayer", "DataAccessObject", "DataAccessObjects",
]


from .types import (
    DALPath, KeyableType
)

from .base import (
    DataAccessLayer, DataAccessObject, DataAccessObjects
)




"""DataAccessLayer Moule Base
"""

__all__ = [
    'DataAccessLayer',
    'DataAccessLayerWithAuth',
    'TableLikeDataAccessLayer',
    'KVLikeDataAccessLayer',
    'QueueLikeDataAccessLayer',
    'PubSubMessage',
    'PubSubLikeDataAccessLayer',
    'DataAccessObject', 
    'DataAccessObjects'
]

import abc
import typing
from typing import Optional as Opt
from .types import DALPath, QueryComLikeType, StrictDALPath, FieldLikeType
from .query_components.modifiers import LimitModifier
from ..utils.enum_ import dump_enum

if typing.TYPE_CHECKING:
    from ..auth import AuthSession
    from ..scheme.field import Field, FieldValueProxy
    from blue_firmament.task.context import ExtendedTaskContext


SchemeTV = typing.TypeVar('SchemeTV', bound="BaseScheme")
FieldValueTV = typing.TypeVar('FieldValueTV')
class DataAccessLayer(abc.ABC):
    """BlueFirmament DataAccessLayer base class.
    """

    def __init_subclass__(
        cls,
        default_path: Opt[StrictDALPath] = None
    ) -> None:
        if default_path is not None:
            cls.__default_path = default_path

    def __init__(self, **kwargs):
        self.__post_init__()

    def __post_init__(self):
        ...

    def dump_path(self, path: Opt[DALPath] = None) -> StrictDALPath:
        """Serialize DALPath to StrictDALPath.

        :returns: StrictDALPath. If `path` is None, returns the default path.
        """
        if path is None:
            return self.__default_path

        return StrictDALPath(tuple(
            dump_enum(i) if i is not None else j 
            for i, j in zip(path, self.__default_path)
        ))

    @property
    def default_path(self) -> StrictDALPath:
        return self.__default_path

    @abc.abstractmethod
    async def close(self):
        raise NotImplementedError("not implemented close method")


class DataAccessLayerWithAuth(DataAccessLayer):

    def __init_subclass__(
        cls,
        force_auth: bool = False,
        **kwargs,
    ):
        cls.__force_auth__ = force_auth
        super().__init_subclass__(**kwargs)

    def __init__(self, auth_session: Opt["AuthSession"] = None, **kwargs):
        if auth_session is None and self.__force_auth__:
            raise ParamsInvalid("auth_session must be provided")
        self._auth_session = auth_session
        super().__init__()

# TODO DataAccessLayerWithConnPool

class TableLikeDataAccessLayer(DataAccessLayer):

    @typing.overload
    @abc.abstractmethod
    async def insert(
        self,
        to_insert: dict,
        path: Opt[DALPath] = None,
        exclude_natural_key: bool = True,
    ) -> dict:
        ...
    @typing.overload
    @abc.abstractmethod
    async def insert(
        self,
        to_insert: SchemeTV,
        path: Opt[DALPath] = None,
        exclude_natural_key: bool = True,
    ) -> SchemeTV:
        ...
    @abc.abstractmethod
    async def insert(
        self,
        to_insert: dict | SchemeTV,
        path: Opt[DALPath] = None,
        exclude_natural_key: bool = True,
    ) -> dict | SchemeTV:
        """Insert a row.

        :param to_insert:
            Data to insert. Can be a dict, model.
        :param path: DALPath.
            Use the model DALPath when None.
        :param exclude_natural_key:
            Exclude natural key fields of the model, defaults to True.

        :returns:
            The inserted row, same type as `to_insert`.
            If `to_insert` is a model, its private fields will be kept.
        """

    @abc.abstractmethod
    async def delete(
        self,
        to_delete: SchemeTV | typing.Type[SchemeTV],
        *query_coms: QueryComLikeType,
        path: Opt[DALPath] = None,
    ) -> None:
        """Delete a row.
        
        :param to_delete: a data model instance or a data model class.
        :param query_coms: filter out rows to be deleted.
            If `to_delete` is a data model instance,
            its primary key will be used as a filter when no filter provided.
        :param path: DALPath.
            Use the data model DALPath when None.

        :raise NoEffect: No row deleted.
        """

    @typing.overload
    @abc.abstractmethod
    async def update(
        self,
        to_update: SchemeTV,
        *query_coms: QueryComLikeType,
        path: Opt[DALPath] = None,
        only_dirty: bool = True,
        exclude_natural_key: bool = True,
    ) -> SchemeTV:
        ...
    @typing.overload
    @abc.abstractmethod
    async def update(
        self,
        to_update: dict,
        *query_coms: QueryComLikeType,
        path: Opt[DALPath] = None,
        only_dirty: bool = True,
        exclude_natural_key: bool = True,
    ) -> dict:
        ...
    @typing.overload
    @abc.abstractmethod
    async def update(
        self,
        to_update: "FieldValueProxy[FieldValueTV]" | FieldValueTV,
        *query_coms: QueryComLikeType,
        path: Opt[DALPath] = None,
        only_dirty: bool = True,
        exclude_natural_key: bool = True,
    ) -> FieldValueTV:
        ...
    @typing.overload
    @abc.abstractmethod
    async def update(
        self,
        to_update: typing.Tuple["Field[FieldValueTV]", FieldValueTV],
        *query_coms: QueryComLikeType,
        path: Opt[DALPath] = None,
        only_dirty: bool = True,
        exclude_natural_key: bool = True,
    ) -> FieldValueTV:
        ...
    @abc.abstractmethod
    async def update(
        self,
        to_update: typing.Union[
            dict, SchemeTV,
            "FieldValueProxy[FieldValueTV]",
            FieldValueTV,  # only for type hint
            typing.Tuple["Field[FieldValueTV]", FieldValueTV]
        ],
        *query_coms: QueryComLikeType,
        path: Opt[DALPath] = None,
        only_dirty: bool = True,
        exclude_natural_key: bool = True,
    ) -> typing.Union[
        dict,
        SchemeTV,
        FieldValueTV,
    ]:
        """Update a row.

        :param to_update: New data.
            Can be:
            - a dict `{'field_a': 2}`, `{Field: 2}`
            - data model instance `MyScheme(field_a=2)`
            - a field value proxy instance `MyScheme.field_a`
            - field and its value `(MyScheme.field_a, 2)`
        :param query_coms:
            Filter out the rows to apply `to_update`.
            Al least one filter must be provided.

            If `to_update` is a data model instance,
            its primary key will be used as a filter when no filter provided.
        :param path: DALPath.
        :param only_dirty: Update only the dirty fields of the data model.
        :param exclude_natural_key: Exclude natural key fields of the model, defaults to True.
        :raises UpdateFailure: 更新失败
        :returns: The updated data.
            If `to_update` is a data model, returns row in data model and keep original
            private fields.
            If `to_update` is field value proxy，returns the new value (without proxy).
            If `to_update` is dict，returns row in dict.
        """

    # TODO add update returns multiple affected records

    @typing.overload
    @abc.abstractmethod
    async def select(
        self,
        to_select: typing.Type[SchemeTV],
        *query_coms: QueryComLikeType,
        path: Opt[DALPath] = None,
        task_context: Opt["ExtendedTaskContext"] = None,
    ) -> typing.Tuple[SchemeTV, ...]:
        ...
    @typing.overload
    @abc.abstractmethod
    async def select(
        self,
        to_select: "Field[FieldValueTV]",
        *query_coms: QueryComLikeType,
        path: Opt[DALPath] = None,
        task_context: Opt["ExtendedTaskContext"] = None,
    ) -> typing.Tuple[FieldValueTV, ...]:
        ...
    @typing.overload
    @abc.abstractmethod 
    async def select(
        self,
        to_select: typing.Iterable[FieldLikeType],
        *query_coms: QueryComLikeType,  # 实际上此时 str, int 不支持
        path: Opt[DALPath] = None,
        task_context: Opt["ExtendedTaskContext"] = None,
    ) -> typing.Tuple[dict, ...]:
        ...
    @abc.abstractmethod
    async def select(
        self,
        to_select: typing.Union[
            typing.Type[SchemeTV], 
            "Field[FieldValueTV]",
            typing.Iterable[FieldLikeType],
        ],
        *query_coms: QueryComLikeType,
        path: Opt[DALPath] = None,
        task_context: Opt["ExtendedTaskContext"] = None,
    ) -> typing.Union[
        typing.Tuple[SchemeTV, ...],
        typing.Tuple[FieldValueTV, ...],
        typing.Tuple[dict, ...]
    ]:
        """Select rows.

        :param to_select: Field(s) to select.
            Can be a data model class, field-like(s).

            If `to_select` is a data model class, all fields of the data model will be selected.
        :param query_coms: Filter out rows to be selected.
            If str, int, an EqFilter of the primary key or field will be created.
        :param path: DALPath.
            If `to_select` is field，use its data model's DAL path.
        :param task_context:
            Task Context injected into return when `to_select` is a data model.
        :raises NotFound: Not a row selected.
        :returns: Selected rows.
            If `to_select` is data model，returns a tuple of data model instances.
            If `to_select` is a field, returns a tuple of field values.
            In other cases, returns a tuple of dicts.

        Examples
        --------
        >>> dao.select(MyScheme, 'primary key value').limit(1)
        # get a MyScheme instance where primary key equals 'primary key value'
        >>> dao.select(MyScheme.field_a, 'primary key value').one()
        # get MySchem.field_a's value using primary key
        >>> dao.select(MyScheme, MyScheme.field_a.equals(value))
        # Get MyScheme using EqFilter on fields other than primary key
        """
        ...

    @typing.overload
    async def select_one(
        self,
        to_select: typing.Type[SchemeTV],
        *query_coms: QueryComLikeType,
        path: Opt[DALPath] = None,
        task_context: Opt["ExtendedTaskContext"] = None,
    ) -> SchemeTV:
        ...
    @typing.overload
    async def select_one(
        self,
        to_select: "Field[FieldValueTV]",
        *query_coms: QueryComLikeType,
        path: Opt[DALPath] = None,
        task_context: Opt["ExtendedTaskContext"] = None,
    ) -> FieldValueTV:
        ...
    @typing.overload
    async def select_one(
        self,
        to_select: typing.Iterable[FieldLikeType],
        *query_coms: QueryComLikeType,
        path: Opt[DALPath] = None,
        task_context: Opt["ExtendedTaskContext"] = None,
    ) -> dict:
        ...
    async def select_one(
        self,
        to_select: typing.Union[
            typing.Type[SchemeTV],
            "Field[FieldValueTV]",
            typing.Iterable[FieldLikeType],
        ],
        *query_coms: QueryComLikeType,
        path: Opt[DALPath] = None,
        task_context: Opt["ExtendedTaskContext"] = None,
    ) -> typing.Union[
        SchemeTV,
        FieldValueTV,
        dict
    ]:
        return (await self.select(
            to_select,
            *query_coms, LimitModifier(1),
            path=path,
            task_context=task_context
        ))[0]

    @abc.abstractmethod
    async def upsert(self):  # TODO
        pass


class KVLikeDataAccessLayer(DataAccessLayer):

    @abc.abstractmethod
    async def get(self, key: str) -> Opt[typing.Any]:
        ...

    @abc.abstractmethod
    async def set(self, key: str, value: typing.Any) -> None:
        ...

class QueueLikeDataAccessLayer(DataAccessLayer):

    def __init_subclass__(
        cls,
        queue_name: Opt[str] = None,
        **kwargs
    ):
        cls.__queue_name__ = queue_name
        super().__init_subclass__(**kwargs)

    def __init__(self, queue_name: Opt[str] = None, **kwargs) -> None:
        if queue_name:
            self.__queue_name__ = queue_name
        super().__init__(**kwargs)

    @abc.abstractmethod
    async def push(self, item: bytes, queue_name: Opt[str] = None) -> None:
        """Push an item to the head of the queue.

        :param item: The item to push.
        :param queue_name: The name of the queue.
            If not provided, the ins_var:queue_name will be used.
        """
        ...

    @abc.abstractmethod
    async def pop(self, queue_name: Opt[str] = None, wait: bool = True) -> bytes:
        """Pop the first item of the queue.

        :param wait: Blocks until an item is available.
        :param queue_name: The name of the queue.
            If not provided, the ins_var:queue_name will be used.
        """
        ...


class PubSubMessage(typing.TypedDict):
    channel: str
    data: bytes

class PubSubLikeDataAccessLayer(DataAccessLayer):
    """

    :cvar __channel_name__: Publish or subscribe to this channel.
    """

    def __init_subclass__(
        cls,
        channel_name: Opt[str] = None,
        **kwargs
    ):
        cls.__channel_name__ = channel_name

        super().__init_subclass__(**kwargs)

    def __init__(self, channel_name: Opt[str] = None, **kwargs) -> None:
        if channel_name:
            self.__channel_name__ = channel_name
        super().__init__(**kwargs)

    @abc.abstractmethod
    async def publish(self, item: bytes, *channel_names: str) -> None:
        """Publish a message to channel(s).

        :param item: The message to publish.
        :param channel_names: Channel names.
            If not provided, the channel_name configured in class level will be used.
        """
        ...

    @abc.abstractmethod
    async def subscribe(self, *channel_names: str) -> None:
        """Subscribe to these channels.

        :param channel_names: Channel names.
            If not provided, the channel_name configured in class level will be used.
        """
        ...

    @abc.abstractmethod
    async def unsubscribe(self, *channel_names: str) -> None:
        """Unsubscribe from these channels.

        :param channel_names: Channel names.
            If not provided, the channel_name configured in class level will be used.
        """
        ...

    @abc.abstractmethod
    async def get_message(self, timeout: int = 0) -> PubSubMessage:
        """Get a message on subscribed channels. (block until a message is received)
        """
        ...

    def listen(self) -> typing.AsyncIterable[PubSubMessage]:
        """Listen to messages on subscribed channels.

        You can iterate over the returned AsyncIterator to get messages.
        """
        ...


class DataAccessObject(
    typing.Generic[SchemeTV]
):
    
    def __init__(
        self,
        dal: DataAccessLayer,
        scheme_cls: type[SchemeTV],
    ) -> None:
        self.__dal = dal
        self.__scheme_cls = scheme_cls

    def select(
        self,
        *query_coms: QueryComLikeType,
        task_context: Opt["ExtendedTaskContext"] = None,
    ):
        if not isinstance(self.__dal, TableLikeDataAccessLayer):
            raise TypeError(f"{self.__dal.__name__} not support TableLike operation")
        return self.__dal.select(
            self.__scheme_cls,
            *query_coms,
            task_context=task_context
        )
        
    def select_fields(
        self,
        to_select: typing.Iterable[FieldLikeType],
        *query_coms: QueryComLikeType,
        task_context: Opt["ExtendedTaskContext"] = None,
    ):
        if not isinstance(self.__dal, TableLikeDataAccessLayer):
            raise TypeError(f"{self.__dal.__name__} not support TableLike operation")
        return self.__dal.select(  
            to_select,
            *query_coms,
            task_context=task_context
        ) 
        
    def select_field(
        self,
        to_select: "Field[FieldValueTV]",
        *query_coms: QueryComLikeType,
        task_context: Opt["ExtendedTaskContext"] = None,
    ):
        """Select a field, returns a tuple of field values.
        """
        if not isinstance(self.__dal, TableLikeDataAccessLayer):
            raise TypeError(f"{self.__dal.__name__} not support TableLike operation")
        return self.__dal.select(
            to_select,
            *query_coms,
            task_context=task_context
        ) 

    def select_one(
        self,
        *query_coms: QueryComLikeType,
        task_context: Opt["ExtendedTaskContext"] = None,
    ):
        if not isinstance(self.__dal, TableLikeDataAccessLayer):
            raise TypeError(f"{self.__dal.__name__} not support TableLike operation")
        return self.__dal.select_one(
            self.__scheme_cls,
            *query_coms,
            task_context=task_context
        ) 
        
    def select_a_field(
        self,
        to_select: "Field[FieldValueTV]",
        *query_coms: QueryComLikeType,
        task_context: Opt["ExtendedTaskContext"] = None,
    ):
        """Select a field, returns a single field value.
        """
        if not isinstance(self.__dal, TableLikeDataAccessLayer):
            raise TypeError(f"{self.__dal.__name__} not support TableLike operation")
        return self.__dal.select_one(
            to_select,
            *query_coms,
            task_context=task_context
        )
    
    def insert(
        self,
        to_insert: SchemeTV,
        exclude_natural_key: bool = True,
    ):
        if not isinstance(self.__dal, TableLikeDataAccessLayer):
            raise TypeError(f"{self.__dal.__name__} not support TableLike operation")
        return self.__dal.insert(
            to_insert=to_insert, 
            exclude_natural_key=exclude_natural_key
        )

    @typing.overload
    async def update(
        self,
        to_update: SchemeTV,
        *query_coms: QueryComLikeType,
        only_dirty: bool = True,
        exclude_natural_key: bool = True,
    ) -> SchemeTV:
        ...
    @typing.overload
    async def update(
        self,
        to_update: FieldValueTV,
        *query_coms: QueryComLikeType,
        only_dirty: bool = True,
        exclude_natural_key: bool = True,
    ) -> FieldValueTV:
        ...
    @typing.overload
    async def update(
        self,
        to_update: typing.Tuple["Field[FieldValueTV]", FieldValueTV],
        *query_coms: QueryComLikeType,
        only_dirty: bool = True,
        exclude_natural_key: bool = True,
    ) -> FieldValueTV:
        ...
    @typing.overload
    async def update(
        self,
        to_update: dict,
        *query_coms: QueryComLikeType,
        only_dirty: bool = True,
        exclude_natural_key: bool = True,
    ) -> dict:
        ...
    def update(
        self,
        to_update: typing.Union[
            dict, SchemeTV,
            FieldValueTV,
            typing.Tuple["Field[FieldValueTV]", FieldValueTV],
        ],
        *query_coms: QueryComLikeType,
        only_dirty: bool = True,
        exclude_natural_key: bool = True,
    ):
        if not isinstance(self.__dal, TableLikeDataAccessLayer):
            raise TypeError(f"{self.__dal.__name__} not support TableLike operation")
        return self.__dal.update(
            to_update,
            *query_coms,
            only_dirty=only_dirty,
            exclude_natural_key=exclude_natural_key,
            path=self.__scheme_cls.dal_path()
        )
    
    def delete(self, 
        *filters: QueryComLikeType,
        to_delete: Opt[SchemeTV] = None,
    ):
        if not isinstance(self.__dal, TableLikeDataAccessLayer):
            raise TypeError(f"{self.__dal.__name__} not support TableLike operation")
        return self.__dal.delete(
            to_delete or self.__scheme_cls,
            *filters
        )


class DataAccessObjects:

    def __init__(
        self,
        auth_session: Opt["AuthSession"] = None
    ) -> None:
        self.__auth_session = auth_session
        self.__dals: dict[type[DataAccessLayer], DataAccessLayer] = {}

    def __call__(self, scheme_cls: typing.Type[SchemeTV]) -> DataAccessObject[SchemeTV]:
        if scheme_cls.__dal__ is None:
            raise ValueError(f"scheme {scheme_cls.__name__} not configured a DAL class")
        
        dal = self.__dals.setdefault(
            scheme_cls.__dal__, 
            scheme_cls.__dal__(auth_session=self.__auth_session) if
            issubclass(scheme_cls.__dal__, DataAccessLayerWithAuth) else
            scheme_cls.__dal__()
        )
        
        return DataAccessObject(
            dal=dal, scheme_cls=scheme_cls
        )

    def is_expired(self) -> bool:
        """Session expired is seen as DAOs expired,"""
        if self.__auth_session:
            return self.__auth_session.is_expired()
        return False


import typing
import abc
from ..task import TaskMetadata

if typing.TYPE_CHECKING:
    from structlog.stdlib import BoundLogger
    from ..core.app import BlueFirmamentApp


class BaseTransporter(abc.ABC):
    """The base class of the transport module.
    """

    def __init__(self, app: "BlueFirmamentApp", name: str = "default") -> None:
        self._app = app
        self._name = name

    def __hash__(self):
        return hash(self._name)

    def __eq__(self, other):
        if isinstance(other, str):
            return self._name == other
        if isinstance(other, BaseTransporter):
            return self._name == other._name
        return False

    def __str__(self):
        return self._name

    @property
    def name(self) -> str:
        """ID of the transporter.
        """
        return self._name

    @property
    def _logger(self) -> "BoundLogger":
        return self._app._logger

    @abc.abstractmethod
    async def start(self):
        """Start listening to tasks
        """

    @abc.abstractmethod
    async def stop(self):
        """Stop listening to tasks
        """

    @staticmethod
    def _parse_task_metadata(raw: typing.Any) -> TaskMetadata:
        if raw is None:
            return TaskMetadata()
        if isinstance(raw, dict):
            return TaskMetadata(**raw)
        raise NotImplementedError(f"Unsupported metadata parsing input type {type(raw)}")




"""Task handler module.
"""

import inspect
import typing
from typing import Optional as Opt

from .._types import PathParamsT, _undefined
from ..exceptions import InternalError
from ..scheme import BaseConverter
from ..scheme.converter import get_converter_from_anno
from . import Task
from .result import TaskResult, Body, EmptyBody, JsonBody
from ..utils.main import call_as_async
from ..utils.typing_ import get_origin, safe_issubclass, is_json_dumpable

if typing.TYPE_CHECKING:
    from .main import BaseTaskContext
    from ..manager import BaseManager


class TaskHandler:
    """BlueFirmament TaskHandler

    Automatically resolve parameters that inner handler (function) needed from
    TaskContext.
    """

    type FunctionKwargsT = typing.Dict[
        str,
        typing.Callable[
            ["BaseTaskContext", PathParamsT], typing.Coroutine,
        ]
    ]
    """Inner handler parameters"""
    type FunctionT = typing.Union[
        typing.Callable[..., typing.Any],
        typing.Callable[..., typing.Awaitable[typing.Any]]
    ]

    def __init__(
        self,
        function: FunctionT,
        manager_cls: Opt[typing.Type["BaseManager"]] = None,
    ):
        """
        :param function:
        :param manager_cls:
            When the inner_handler is a manager method
            (excludes classmethod, staticmethod).
        """
        self.__function = function
        self.__method_manager_cls = manager_cls

        # parse handler kwargs
        self.__handler_kwargs: TaskHandler.FunctionKwargsT =\
            self._parse_handler_kwargs(self.__function)

    @property
    def function(self) -> FunctionT:
        return self.__function

    def set_manager_cls(self, manager_cls: typing.Type["BaseManager"]):
        """Set inner handler's manager class if it's a manager method.

        Useful for manager class created after task handler created.
        E.g. use :meth:`blue_firmament.task.task` to decorate a manager method.
        """
        self.__method_manager_cls = manager_cls

    @staticmethod
    def get_param_getter(name: str, converter: BaseConverter, safe: bool = True):
        """Get a getter resolves parameter from
        task parameters or path parameters.
        """
        async def getter(tc: "BaseTaskContext", path_params: PathParamsT):
            val = path_params.get(name, _undefined)
            if val is _undefined:
                val = await tc._task.parameters.get(name, _undefined)
                if val is _undefined:
                    if not safe:
                        raise ValueError(f'{name} not found in task parameters or path parameters')
                    else:
                        return _undefined
            return converter(val, _task_context=tc)

        return getter

    @classmethod
    def _parse_handler_kwargs(cls, handler: FunctionT) -> FunctionKwargsT:
        """
        Rationale
        ---------
        - 此处假设处理器的签名是静态的，所以可以在路由记录实例化时解析参数备用，无需每次调用处理器之前重新解析
        - ``__call__`` 等调用处理器时将会使用解析的结果来自动注入参数

        Behaviour
        ----------
        参数解析规则
        ^^^^^^^^^^^^^^^
        - 跳过这些名称的参数
            - `self`, `cls`
        - 基于类型解析系统信息
            - `Response`, `Request`
        - 基于名称解析请求体参数（只能有一个，名称为 `HANDLER_BODY_KW` ）
        - 基于名称和类型解析路径参数、查询参数
            - 路径参数优先于查询参数
            - 名称匹配但类型不匹配会导致 `ValueError` 异常

        Dependency
        -----------
        - :meth:`scheme.converter.get_converter_from_anno`
        - `utils.type.get_origin`

        Returns
        -------
        返回一个字典，键为处理器的参数名称，值为该参数的获取器。

        参数获取器接收 :class:`blue_firmament.transport.context.RequestContext` 作为参数，从中解析出本参数需要的值。
        """
        handler_params_sig = inspect.signature(handler).parameters
        kwargs: TaskHandler.FunctionKwargsT = {}

        for name, param_sig in handler_params_sig.items():
            if name in ('self', 'cls'):
                continue

            anno = get_origin(param_sig.annotation)
            converter = get_converter_from_anno(param_sig.annotation)

            if safe_issubclass(anno, Task):
                kwargs[name] = lambda tc, _: tc._task
                continue
            elif safe_issubclass(anno, TaskResult):
                kwargs[name] = lambda tc, _: tc._task_result
                continue

            kwargs[name] = cls.get_param_getter(name, converter)

        return kwargs

    async def __call__(
        self, *,
        task_context: 'BaseTaskContext',
        path_params: PathParamsT,
    ) -> Body:
        """Run the inner handler.

        Behaviour
        ---------
        - 自动传递参数：基于实例化时解析的 `RequestHandlerKwargs`，从环境中解析出参数
        - 自动处理异步/同步函数：通过 `inspect.iscoroutinefunction` 判断函数是否为异步函数
        - 自动处理返回值：处理器的返回值会被解析到响应对象中
        """
        # get kwargs
        kwargs = {}
        for name, getter in self.__handler_kwargs.items():
            value = await getter(task_context, path_params)
            if value is not _undefined:
                kwargs[name] = value

        # get args
        args = []
        # [self] don't add other arg parser before this one
        if self.__method_manager_cls:
            manager = self.__method_manager_cls(task_context=task_context)
            args.append(manager)

        # call handler
        result = await call_as_async(self.__function, *args, **kwargs)

        # process result
        # TODO process result correctly
        if isinstance(result, Body):
            return result
        elif result is None:
            return EmptyBody()
        elif is_json_dumpable(result):
            return JsonBody(result)
        else:
            raise InternalError("handler returns invalid value")





import abc
import datetime
import typing
from ..utils.datetime_ import get_datetimez
from ..data.settings.base import get_setting as get_base_setting

if typing.TYPE_CHECKING:
    from blue_firmament.task import Task


SFValueTV = typing.TypeVar("SFValueTV")
class SessionField(typing.Generic[SFValueTV], abc.ABC):
    """会话字段类（抽象类）

    是状态的基本存储单位

    声明周期
    ^^^^^^^
    - 创建：在连接的第一个请求到达时初始化
    - 销毁
    - 重新创建
    - 刷新

    Example
    ^^^^^^^
    ```python
    class MySessionField(SessionField):
    
        def __init__(self):
            ...    
            super().__init__(value)
    ```
    """

    def __init__(self, value: SFValueTV):
        self.__updated_at = get_datetimez()
        self.__value: SFValueTV = value
    
    @property
    def updated_at(self):
        return self.__updated_at

    @property
    def value(self):
        return self.__value
    @value.setter
    def value(self, value: SFValueTV): 
        self.__value = value

    def destroy(self) -> None:
        pass

    def is_expired(self) -> bool:
        """Tell whether this field is expired.

        Always False by default, subclasses should override this method.
        """
        return False

    def refresh(self) -> None:
        pass

    @classmethod
    def from_task(cls, task: 'Task') -> typing.Self:
        raise NotImplementedError("this field not support from_task method")

    
class Session(abc.ABC):
    """BlueFirmament Session base class
    """

    INACTIVE_THRESHOLD = 300
    """unit: seconds"""
    REMOVE_BATCH_SIZE = 10
    """how many sessions to remove once session pool is full"""
    SESSION_POOL_MAX = 1000
    """max session cached in session pool"""

    __session_pool__: dict[str, typing.Self] = {}
    # TODO use remote storage to share session and save local memory (but performance cost?)
    """session cache pool"""
    __fields__: tuple[str, ...] = ()
    """all field's name"""

    def __init__(self, _id: str, /, **fields: SessionField) -> None:
        # metadata (not session fields, raw value)
        self.__id: str = _id
        self.__last_used_at: datetime.datetime = get_datetimez()

        self.__init_fields__(**fields)

    @abc.abstractmethod
    def __init_fields__(self, **fields: SessionField):
        ...

    @property
    def id(self):
        """Session ID"""
        return self.__id

    @property
    def last_used_at(self):
        return self.__last_used_at

    def update_last_used_at(self):
        self.__last_used_at = get_datetimez()

    def is_expired(self) -> bool:
        """Any field expired makes session expired.

        Will refresh expired fields first, if still expired, return True.
        """
        for field_name in self.__fields__:
            field = getattr(self, f"_{self.__class__.__name__}__{field_name}")
            if isinstance(field, SessionField):
                if field.is_expired():
                    field.refresh()
                    if field.is_expired():
                        return True
        return False

    def is_inactive(self) -> bool:
        """A span of time has passed since last use.
        """
        return get_datetimez() - self.__last_used_at > datetime.timedelta(seconds=self.INACTIVE_THRESHOLD)

    @classmethod
    def cleanup_sessions(cls) -> None:
        """Delete expired, inactive sessions from session pool.

        TODO run periodically.
        """
        to_delete = []

        for _id, session in cls.__session_pool__.items():
            if session.is_expired() or session.is_inactive():
                to_delete.append(_id)

        if (len(cls.__session_pool__) - len(to_delete)) >= cls.SESSION_POOL_MAX:
            # randomly remove a batch of sessions
            to_delete_len = len(to_delete)
            for key in cls.__session_pool__.keys():
                to_delete.append(key)
                if (len(to_delete) - to_delete_len) >= cls.REMOVE_BATCH_SIZE:
                    break

        for _id in to_delete:
            del cls.__session_pool__[_id]

    @classmethod
    def upsert(
        cls,
        _id: str,
        fields_getter: typing.Callable[[], dict[str, SessionField]],
    ) -> typing.Self:
        """

        :param _id: Session ID
        :param fields_getter: function returns fields for creating a new session.

        If session expire, recreate.
        """
        session = cls.__session_pool__.setdefault(_id, cls(_id, **fields_getter()))
        if session.is_expired():
            del cls.__session_pool__[_id]
            session = cls(_id, **fields_getter())
        session.update_last_used_at()
        return session
    
    @classmethod
    @abc.abstractmethod
    def from_task(cls, task: 'Task') -> typing.Self:
        """Create a session from a task
        """
        ...




"""BlueFirmament Scheme Field module
"""

__all__ = [
    "FieldValueProxy",
    "Field",
    "PrivateField",
    "FieldTV",
    "CompositeField",
    "field",
    "private_field",
    "get_default",
    "dump_field_name",
]

import functools
import typing
from typing import Optional as Opt
from .._types import Undefined, _undefined
from ..utils.typing_ import safe_issubclass
from blue_firmament.dal.query_components.filters import (
    ContainsFilter, EqFilter, NotEqFilter,
    InFilter, IsFilter
)
from ..dal.query_components.modifiers import OrderModifier
from blue_firmament.dal.query_components.operators import NotOperator
from .converter import BaseConverter, get_converter_from_anno

if typing.TYPE_CHECKING:
    from .main import BaseScheme
    from .validator import BaseValidator


FieldValueTV = typing.TypeVar('FieldValueTV')
class FieldValueProxy(typing.Generic[FieldValueTV]):
    """字段值代理对象

    字段值必然是原生值，不能是用户自定义类

    - 检测对可变对象的修改
    - 获得所属的字段实例

    Implementation
    --------------
    透明代理
    ^^^^^^^^^^
    - 在下方通过 for 循环注册了一堆特殊方法
    - `__bool__` 特殊处理
    """

    def __init__(self, 
        obj: FieldValueTV, 
        modified: typing.Callable,
        field: "Field[FieldValueTV]",
        scheme: "BaseScheme"
    ) -> None:

        self._obj: FieldValueTV = obj
        self._modified = modified
        self._field: "Field[FieldValueTV]" = field
        self._scheme: "BaseScheme" = scheme

    @property
    def scheme(self): return self._scheme
    
    @property
    def field(self): return self._field

    @property
    def obj(self) -> FieldValueTV:
        """获取原始对象

        这个对象是不可变的
        """
        return self._obj

    @staticmethod
    def _get_obj_dunder_method_caller(name: str) -> typing.Callable:

        def dunder_method_caller(self: typing.Self, *args, **kwargs):
            return getattr(self._obj, name)(*args, **kwargs)

        return dunder_method_caller

    def __bool__(self):

        try:
            return getattr(self._obj, '__bool__')()
        except AttributeError:
            return bool(self._obj)

    def __getattr__(self, name: str) -> typing.Any:
        
        attr = getattr(self._obj, name)
        
        # 代理方法
        if callable(attr):
            @functools.wraps(attr)
            def wrapper(*args, **kwargs):
                res = attr(*args, **kwargs)
                self._modified()  # 字段内部修改
                # TODO 并不是所有方法都会导致值的修改，这也没办法
                return res
            return wrapper

        return attr
    
    def __setattr__(self, name: str, value: typing.Any) -> None:
        
        if name in ('_obj', '_modified', '_field', '_scheme'):
            super().__setattr__(name, value)
        else:
            setattr(self._obj, name, value)

    @staticmethod
    def dump(value: FieldValueTV | "FieldValueProxy[FieldValueTV]") -> FieldValueTV:
        if isinstance(value, FieldValueProxy):
            return value.obj
        return value


for i in (
    'add', 'sub', 'mul', 'truediv', 'floordiv', 'mod', 'pow',
    'eq', 'ne', 'lt', 'le', 'gt', 'ge',
    'len', 'getitem', 'setitem', 'contains',
    'int', 'float', 'str', 'repr',
    'hash', 'call', 'iter',
    'next', 'reversed', 'abs', 'round',
    'and', 'or', 'xor', 'invert',
    'lshift', 'rshift'
):
    setattr(
        FieldValueProxy, f'__{i}__', 
        FieldValueProxy._get_obj_dunder_method_caller(f'__{i}__')
    )

@typing.runtime_checkable
class FieldValueProtocol(typing.Protocol[FieldValueTV]):

    __field__: "Field[FieldValueTV]"
    __scheme__: "BaseScheme"


class Field(typing.Generic[FieldValueTV]):

    """Scheme Field of BlueFirmament.

    Features
    --------
    Get DAL Filter
    ^^^^^^^^^^^^^^
    >>> Field[int](name='_id').equals(1)
    EqFilter(field='_id', value=1)

    Validator
    ^^^^^^^^^
    See :doc:`/design/scheme/validator`

    Inherit
    ^^^^^^^
    - Make sure your `__init__` accepts the same parameters as `Field` or use
    `**kwargs` to achieve that. (Set your parameters in kwargs to avoid conflict)
      Or set your parameters in `_post_init` instead of
    overriding `__init__`.
    """

    def __init__(
        self, 
        default: typing.Union[Undefined, FieldValueTV] = _undefined, 
        default_factory: Opt[typing.Callable[[], FieldValueTV]] = None,
        vtype: Undefined | typing.Type[FieldValueTV] = _undefined,
        name: Opt[str] = None,
        in_scheme_name: Opt[str] = None,
        scheme_cls: Opt[typing.Type["BaseScheme"]] = None,
        is_key: bool = False,
        is_key_natural: bool = False,
        is_foreign_key: bool = False,
        converter: Opt[BaseConverter[FieldValueTV]] = None,
        validators: Opt[typing.Iterable["BaseValidator"]] = None,
        is_partial: bool = False,
        dump_flags: Opt[set[str]] = None,
        init: bool = True,
    ):
        """
        :param default:
            Immutable value used when set value not provided(``_undefined``).
        :param default_factory:
            Callable to get mutable value as default value.
        :param vtype: 
            Field value type.
        :param name:
            Field name in DAL.
        :param in_scheme_name:
            Field name in scheme.
        :param scheme_cls:
            Scheme class this field attached on.
        :param converter:
            See :doc:`/design/scheme/converter`
        :param validators:
            See :doc:`/design/scheme/validator`
        :param is_partial:
            If True, field value left undefined if not provided, and
            default value will not be applied.
        :param dump_flags:
            Flags control whether this field should be dumped when
            dumping scheme.
        :param is_key: 
            This field is a key
        :param is_key_natural:
            If this field is a key, it is managed by DataSource (False for surrogate)
        :param is_foreign_key:
            If you want to mark a primary or composite key, use KeyField instead
        :param init: 
            see `Dataclass field specifier parameters <https://typing.python.org/en/latest/spec/dataclasses.html#field-specifier-parameters>`_
        """
        self.__name = name
        self.__in_scheme_name = in_scheme_name or name
        self.__scheme_cls = scheme_cls
        self.__default = default
        self.__default_factory = default_factory
        self.__vtype = vtype
        self.__is_key = is_key
        self.__is_key_natural = is_key_natural
        self.__is_foreign_key = is_foreign_key
        self.__converter: BaseConverter | None = converter
        self.__validators: typing.List["BaseValidator"] = list(validators or [])
        self.__is_partial = is_partial
        self.__dump_flags = dump_flags or set()
        self.__init = init

    # FIXME all default must be negative, or default value will override original value even if \
    #  no new value is set
    def fork(
        self,
        default: Undefined | FieldValueTV = _undefined,
        default_factory: Opt[typing.Callable[[], FieldValueTV]] = None,
        vtype: Undefined | typing.Type[FieldValueTV] = _undefined,
        name: Opt[str] = None,
        in_scheme_name: Opt[str] = None,
        scheme_cls: Opt[typing.Type["BaseScheme"]] = None,
        is_key: bool = False,
        is_key_natural: bool = False,
        is_foreign_key: bool = False,
        converter: Opt[BaseConverter[FieldValueTV]] = None,
        fork_validators: bool = False,
        is_partial: Opt[bool] = None,
        dump_flags: Opt[set[str]] = None,
        init: Opt[bool] = None,
    ) -> typing.Self:
        return self.__class__(
            default=default if default is not _undefined else self.__default,
            default_factory=default_factory or self.__default_factory,
            name=name or self.__name,
            in_scheme_name=in_scheme_name or self.__in_scheme_name,
            scheme_cls=scheme_cls or self.__scheme_cls,
            is_key=is_key or self.__is_key,
            is_key_natural=is_key_natural or self.__is_key_natural,
            is_foreign_key=is_foreign_key or self.__is_foreign_key,
            converter=converter or self.__converter,
            validators=self.__validators if fork_validators else None,
            is_partial=is_partial or self.__is_partial,
            dump_flags=dump_flags or self.__dump_flags,
            init=init or self.__init,
            vtype=vtype if vtype is not _undefined else self.__vtype
        )
    
    def __hash__(self) -> int:
        """哈希值

        in_scheme_name or name
        """
        return hash(self.__in_scheme_name or self.__name)
    
    def __eq__(self, value) -> bool:

        """Make BlueFirmamentField(name='name') == 'name'
        """
        if isinstance(value, str):
            return self.__name == value or self.__in_scheme_name == value
        elif isinstance(value, Field):
            # name, vtype
            return self.__name == value.name and self.value_type == value.value_type
        return False

    @property
    def name(self) -> str: 
        if self.__name is None:
            raise ValueError('Field name is not defined')
        return self.__name
    
    @property
    def in_scheme_name(self) -> str:
        if self.__in_scheme_name is None:
            raise ValueError('Field in_scheme_name is not defined')
        return self.__in_scheme_name
    
    def is_key(self) -> bool: 
        return self.__is_key
    def is_key_natural(self) -> bool:
        return self.__is_key_natural

    def _set_name(self, value: str, no_raise: bool = False) -> None:
        """设置字段名称

        如果已经设置过名称，则抛出错误；如果不想抛出错误，则传入 ``no_raise`` 参数为 ``True``
        """
        if self.__name is not None:
            if no_raise:
                return None
            raise ValueError('Field name is immutable')
        self.__name = value

    def _set_in_scheme_name(self, value: str, no_raise: bool = False) -> None:
        """设置字段在数据模型中的名称

        如果已经设置过名称，则抛出错误
        """
        if self.__in_scheme_name is not None:
            if no_raise:
                return None
            raise ValueError('Field in_scheme_name is immutable')
        self.__in_scheme_name = value

    @property
    def vtype(self) -> typing.Type[FieldValueTV]:
        if self.__vtype is _undefined:
            raise ValueError(f"vtype is not set on field {self.name}")
        return self.__vtype
    
    @property
    def is_partial(self) -> bool:
        return self.__is_partial

    @property
    def scheme_cls(self) -> typing.Type["BaseScheme"]: 
        if self.__scheme_cls is None:
            raise ValueError('Field scheme is not defined')
        return self.__scheme_cls
    
    @property
    def dump_flags(self) -> set[str]:
        return self.__dump_flags

    @dump_flags.setter
    def dump_flags(self, value: set[str]) -> None:
        self.__dump_flags = value

    def _set_scheme_cls(self, 
        scheme_cls: Opt[typing.Type["BaseScheme"]],
        no_raise: bool = False,
        force: bool = False
    ) -> None:
        """设置字段所属的数据模型

        如果已经设置过数据模型，则抛出错误

        :param force: 是否强制设置 \n
            建议使用 :meth:`fork` 来设置值（但元类是特例）
        """
        if self.__scheme_cls is not None and not force:
            if no_raise: return None
            raise ValueError('Field scheme is immutable')
        self.__scheme_cls = scheme_cls

    @property
    def init(self) -> bool: return self.__init

    def _set_converter(self, 
        converter: BaseConverter[FieldValueTV],
        safe: bool = True
    ) -> None:

        """
        :param converter:

        :raise ValueError: Conveter has been set
        """
        if self.__converter is not None:
            if not safe:
                raise ValueError('converter is immutable')
            else:
                return
        self.__converter = converter

    def _set_converter_from_anno(self, 
        annotation: typing.Type[FieldValueTV],
        safe: bool = True
    ) -> None:

        """从类型注解设置转换器

        - 支持 `BlueFirmamentField[type]`
        - 支持 `BlueFirmamentFieldSublcass` （必须是直接子类，不能是孙类）
        - 用户不应当调用

        :param annotation: 类型注解
        :param safe: 已经设置时是否不报错
        :raise ValueError: 如果已经设置过转换器
        """
        if self.__converter is not None:
            if not safe:
                raise ValueError('converter is immutable')
            else:
                return

        if typing.get_origin(annotation) is self.__class__:
            annotation = typing.get_args(annotation)[0]
        else:
            if safe_issubclass(annotation, Field):
                annotation = annotation.__orig_bases__[0]  # type: ignore[attr-defined]

        self.__converter = get_converter_from_anno(annotation)

    def _add_validator(self, validator: "BaseValidator") -> None:

        """添加校验器

        :param validator: 校验器

        - 用户不应当调用
        """
        self.__validators.append(validator)

    @property
    def value_type(self) -> typing.Type[FieldValueTV]:
        
        """获取字段值类型

        从校验器中获取
        """
        if self.__converter:
            return self.__converter.type
        else:
            raise ValueError('Field value type is not defined')

    def load_val(self, value: typing.Any) -> FieldValueTV:
        if value is _undefined:
            return self.default_value
        else:
            return self.convert(value)
    
    def dump_val_to_str(self, value: FieldValueTV):
        """Dump field value to string
        """
        if value is not _undefined:
            return self.converter.dump_to_str(value)
        else:
            return _undefined.value

    def dump_val_to_jsonable(self, value: FieldValueTV):
        """Dump field value to jsonable types
        """
        if value is not _undefined:
            return self.converter.dump_to_jsonable(value)
        else:
            return _undefined.value

    @property
    def converter(self) -> BaseConverter:
        if self.__converter is None:
            raise ValueError('converter is not defined on field %s' % self.in_scheme_name)
        return self.__converter

    @property
    def default_value(self) -> FieldValueTV:

        """默认值

        - 默认值工厂优先于默认值

        :raise ValueError: If no default value provided.
        """
        if self.__default_factory:
            return self.__default_factory()
        elif self.__default is not _undefined:
            return self.__default
        else:
            raise ValueError('No default value provided for field %s' % self.in_scheme_name)
    
    def convert(self, value: typing.Any) -> FieldValueTV:

        """转换字段值

        返回转换后的值（如果无法处理会抛出错误）；
        没有定义转换器则直接返回值
        """
        if self.__converter:
            return self.__converter(value)
        else:
            return value
        
    def validate(self, 
        value: typing.Any,
        scheme_ins: Opt["BaseScheme"] = None,
    ) -> None:
        """Run validator bind to this field.

        :raises ValueError: If invalid.
        """
        for validator in self.__validators:
            validator(value, scheme_ins=scheme_ins)
        
    def equals(self, value: typing.Any) -> EqFilter:
        """该字段等于该值的筛选器
        """
        return EqFilter(self, value)
    
    def contains(self, *value: typing.Any) -> ContainsFilter:
        """该字段包含所有元素的筛选器
        """
        return ContainsFilter(self, *value)
    
    def not_equals(self, value: typing.Any) -> NotEqFilter:
        """该字段不等于该值的筛选器
        """
        return NotEqFilter(self, value)
    
    def in_(self, value: typing.Iterable[typing.Any]) -> InFilter:
        """该字段在该列表中的筛选器
        """
        return InFilter(self, value)
    
    def not_in_(self, value: typing.Iterable[typing.Any]) -> tuple[NotOperator, InFilter]:
        return (NotOperator(), self.in_(value))

    def is_(self, value: bool | None) -> IsFilter:
        return IsFilter(self, value)

    def is_not(self, value: bool | None) -> tuple[NotOperator, IsFilter]:
        return (NotOperator(), IsFilter(self, value))
    
    def order_by(self, *, desc: bool = False) -> OrderModifier:

        """按该字段进行排序的修改器
        """
        return OrderModifier(self, desc=desc)

    @typing.overload
    def __get__(self, instance: None, owner) -> typing.Self:
        ...

    @typing.overload
    def __get__(self, instance: "BaseScheme", owner) -> FieldValueTV:
        # 实际上是 FieldValueProxy[FieldValueType]
        ...
    
    def __get__(self, instance: Opt["BaseScheme"], owner) \
        -> "Field" | FieldValueTV:

        if instance is None:
            return self
        return typing.cast(FieldValueTV, instance._get_value(self))
        
    def __set__(self, instance: "BaseScheme", value: FieldValueTV) -> None:
        """
        .. versionchanged:: 0.1.2
            if initialized, set to undefined will change nothing
        """
        initialized = self.in_scheme_name in instance.__field_values__

        # convert value
        if value is _undefined:
            if initialized:
                return  # remain value unchanged

            instance._mark_unset(self.in_scheme_name)
            try:
                value = self.default_value
            except ValueError as e:
                if self.__is_partial or instance.__partial__:
                    instance._set_value(self, _undefined)
                    return
                else:
                    raise e
        else:
            value = self.convert(value)

        # validate value
        self.validate(value, scheme_ins=instance)

        if instance.__proxy__:
            value_ = self._proxy_value(value, instance)
        else:
            value_ = value
        
        instance._set_value(self, value_)  # save

        # if already initialized, mark as dirty
        if initialized:
            instance._mark_dirty(self.in_scheme_name)

    def _proxy_value(self, 
        value: FieldValueTV, instance: "BaseScheme"
    ) -> FieldValueProxy[FieldValueTV]:
        
        if not isinstance(value, FieldValueProxy):
            # 避免循环代理
            return FieldValueProxy(
                value, 
                lambda: instance._mark_dirty(self.in_scheme_name),
                self,
                instance
            )
        return value  # 已经是代理对象了

    def dump_to_dict(self, value: FieldValueTV) -> dict[str, FieldValueTV]:

        """获取字段及其值组成的字典

        :param value: 字段值；会先由校验器处理
        """
        return {
            self.name: self.convert(value)
        }

    def dump_to_scheme(self) -> typing.Type["BaseScheme"]:

        """
        根据字段实例创建数据模型

        这个数据模型只包含本字段
        """

        exec_namespace = {
            "BaseScheme": BaseScheme,
            "FieldIns": self,
        }
        exec_result = {}

        class_sig = "class AnonymousScheme(BaseScheme):\n"
        class_body = f"    {self.name} = FieldIns\n"

        exec(class_sig + class_body, exec_namespace, exec_result)
        return exec_result["AnonymousScheme"]


T = typing.TypeVar('T')
def field(
    default: T | Undefined = _undefined, 
    default_factory: Opt[typing.Callable[[], T]] = None,
    name: Opt[str] = None,
    is_key: bool = False,
    is_natural_key: bool = False,
    is_foreign_key: bool = False,  # TODO no needed, remove
    converter: Opt[BaseConverter] = None,
    validators: Opt[typing.Iterable['BaseValidator']] = None,
    is_partial: bool = False,
    dump_flags: Opt[set[str]] = None,
    init: bool = True
):
    return Field[T](
        default=default, 
        default_factory=default_factory, 
        name=name, 
        is_key=is_key or is_natural_key,
        is_key_natural=is_natural_key,
        is_foreign_key=is_foreign_key, 
        converter=converter,
        validators=validators,
        is_partial=is_partial,
        dump_flags=dump_flags,
        init=init
    )


class PrivateField[FieldValueType](Field[FieldValueType]):

    """碧霄私有字段

    - 不会被外部访问？（setting 那边的实践有问题）
    - 不会被序列化
    - 不会与数据访问层交互（无需代理）
    - 不会被校验
    - 实例化时不是必须的
    """
    
    @property
    def name(self): raise ValueError('Private field name is forbidden')
    
    def __set__(self, instance: "BaseScheme", value: FieldValueType) -> None:
        try:
            if value is _undefined:
                value = self.default_value
        except ValueError:
            value = _undefined
        instance._set_value(self, value)

def private_field(
    default: T | Undefined = _undefined,
    default_factory: Opt[typing.Callable[[], T]] = None,
):
    
    return PrivateField[T](
        default=default, 
        default_factory=default_factory, 
    )


SchemeTV = typing.TypeVar("SchemeTV", bound="BaseScheme")
class CompositeField(
    Field[SchemeTV],
    typing.Generic[SchemeTV], 
):
    """

    - Enable partial for composite field makes sub scheme partial.
    """
    
    @property
    def name(self): 
        raise ValueError("CompositeField don't has a name")
    def _set_name(self, value: str, no_raise: bool = False) -> None:
        if no_raise:
            return None
        raise ValueError("CompositeField can't have a name")
    
    def dump_val_to_jsonable(self, value: SchemeTV) -> typing.Dict[str, typing.Any]:
        return value.dump_to_dict(jsonable=True)
    
    @property
    def sub_fields(self) -> typing.Iterable[Field]:
        return self.vtype.__fields__.values()
    
    @property
    def _sub(self) -> typing.Type[SchemeTV]:
        return self.vtype


def get_default(field: Field[T]) -> T:

    '''Get field's default value.

    :raise ValueError: 
    '''
    return typing.cast(Field[T], field).default_value

def dump_field_name(field: Field | str, in_scheme: bool = False) -> str:

    if isinstance(field, Field):
        return field.name if not in_scheme else field.in_scheme_name
    return field


FieldTV = typing.TypeVar('FieldTV', bound=Field)
"""字段类型变量"""




import logging
import typing
from typing import Optional as Opt, Annotated as Anno, Literal as Lit
from ...setting import Setting, make_setting_singleton
from ...scheme import private_field


class LogSetting(Setting):

    __disable_log__ = True

    _setting_name = private_field("log")
    
    log_level: int = logging.INFO
    log_file: Opt[str]  = None
    cache_logger_on_first_use: bool = False


get_setting, set_setting = make_setting_singleton(LogSetting())




"""BlueFirmament Dependency Injection framework
"""

import inspect
import typing
from typing import Optional as Opt, Annotated as Anno, Literal as Lit

from .utils.main import call_as_sync
from .utils.inspect_ import args_to_kwargs_by_sig


class DependencyInjector:

    """
    :ivar pargs: Position only arguments
    :ivar kargs: Keyword or position arguments
    """

    class ArgGetterT(typing.Protocol):
        
        """A callable use to get the argument value.
        """

        def __call__(self,
            args: typing.Tuple[typing.Any], 
            kwargs: typing.Dict[str, typing.Any]
        ) -> typing.Any:
            
            """
            :param args: Positional arguments.
            :param kwargs: Keyword arguments.
            """
            ...

    class DynamicArgGetterT(typing.Protocol):

        """A callable use to get argument getter
        """

        def __call__(self, anno: typing.Type) -> "DependencyInjector.ArgGetterT":
            ...
    
    __getters_by_name__: typing.Mapping[str, ArgGetterT] = {}
    """A dictionary of argument name to getter function."""
    __getters_by_anno__: typing.Mapping[typing.Type, ArgGetterT] = {}
    """A dictionary of argument annotation to getter function."""
    __dgetters_by_name__: typing.Mapping[str, DynamicArgGetterT] = {}
    """A dictionary of argument name to dynamic getter function."""

    def __init__(self, 
        callable: typing.Callable,
        getters_by_name: Opt[typing.Mapping[str, ArgGetterT]] = None,
        getters_by_anno: Opt[typing.Mapping[typing.Type, ArgGetterT]] = None,
        dgetters_by_name: Opt[typing.Mapping[str, DynamicArgGetterT]] = None,
    ) -> None:

        """
        :param callable: The callable to be injected.

            - Can be async function. (run synchronously)
        :param getters_by_name: Override class level getters by name.
        :param getters_by_anno: Override class level getters by annotation.
        """

        self.__callable = callable
        if getters_by_name:
            self.__getters_by_name__ = getters_by_name
        if getters_by_anno:
            self.__getters_by_anno__ = getters_by_anno
        if dgetters_by_name:
            self.__dgetters_by_name__ = dgetters_by_name
        
        self.__pargs, self.__kargs = self._parse_args(callable)

    @property
    def name(self) -> str:
        return self.__callable.__qualname__

    def __call__(self, *rargs, **rkwargs) -> typing.Any:
        
        """Call the callable with the arguments.

        If argument has no getter, use raw value from args or kwargs.

        :param rargs: Positional only arguments.
        :param rkwargs: Keyword or positional arguments.

            Keyword arg that is not in the signature will be ignored.
        """
        args = list()
        kwargs = dict()
        for i, v in enumerate(self.__pargs):
            if v:
                args.append(v(rargs, rkwargs))
            else:
                try:
                    args.append(rargs[i])
                except IndexError:
                    continue
        for k, v in self.__kargs.items():
            if v:
                kwargs[k] = v(rargs, rkwargs)
            else:
                try:
                    kwargs[k] = rkwargs[k]
                except KeyError:
                    continue

        return call_as_sync(self.__callable, *args, **kwargs)

    def _parse_args(self, callable: typing.Callable) -> typing.Tuple[
        typing.Tuple[Opt[ArgGetterT], ...],
        typing.Dict[str, Opt[ArgGetterT]]
    ]:
        """
        
        """
        parameters_sig = inspect.signature(callable).parameters
        pargs = []
        kargs = {}
        for name, param in parameters_sig.items():
            getter = self._get_arg_getter(name, param.annotation)
            if param.kind == inspect.Parameter.POSITIONAL_ONLY:
                pargs.append(getter)
            elif param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                kargs[name] = getter

        return tuple(pargs), kargs
    
    def _get_arg_getter(self,
        name: str,
        anno: typing.Type,
    ) -> typing.Optional[ArgGetterT]:
        
        """Get the argument getter function.

        Priority:
        1. Getters by name
        2. Getters by annotation
        3. Dynamic getters by name

        :return: The argument getter function or None.
        """
        res = self.__getters_by_name__.get(name)
        if not res:
            res = self.__getters_by_anno__.get(anno)
        if not res:
            res_ = self.__dgetters_by_name__.get(name)
            if res_:
                res = res_(anno)

        return res

    def pargs_to_kwargs(self, *pargs) -> typing.Dict[str, typing.Any]:

        """Convert positional arguments to keyword arguments.

        :param pargs: Positional arguments.
        :return: Keyword arguments.
        """
        return args_to_kwargs_by_sig(self.__callable, *pargs)


"""BlueFirmament Scheme's Validator module.

:doc:`/design/scheme/validator`
"""

__all__ = [
    "BaseValidator",
    "FieldValidator",
    "field_validator",
]

import abc
import copy
import typing
from typing import Optional as Opt, Annotated as Anno, Literal as Lit

from .._types import Undefined, _undefined
from ..utils.main import call_as_sync

if typing.TYPE_CHECKING:
    from .field import Field
    from .main import BaseScheme
    from ..log.main import LoggerT


T = typing.TypeVar("T")
BaseSchemeTV = typing.TypeVar("BaseSchemeTV", bound="BaseScheme", contravariant=True)
class BaseValidator(abc.ABC, typing.Generic[T]):

    """Base class for validators.
    """

    @abc.abstractmethod
    def _get_logger(self, *args, **kwargs) -> "LoggerT":

        """Validator level logger.

        Logger binded to validator context
        """

    @abc.abstractmethod
    def __call__(self, value: T, **kwargs) -> None:

        """Validate the value.

        :param value: The value to validate.
        :raises ValueError: If the value is invalid.
        """


class FieldValidator(BaseValidator[T], typing.Generic[T]):

    """Validator for a field.

    This validator will be added to the field once instantiated.
    """

    class InstanceMethodFunc(typing.Protocol[BaseSchemeTV]):
        def __call__(_, self: BaseSchemeTV, value: typing.Any) -> typing.Union[
            None, typing.Coroutine[None, None, None]
        ]:
            ...
            
    FuncT = typing.Union[
        InstanceMethodFunc
    ]
    """Function types that FieldValidator can accept.
    """
    ModeT = Lit["init_assign", "after", "assign"]

    def __init__(self, 
        field: "Field",
        func: FuncT,
        mode: ModeT = "after",
    ) -> None:
        
        """
        :param field: The field to validate.
        :param func: The function to call to validate the field.
        :param mode: 
            init_assign: validate only when assigning during scheme instantiating.
            after: validate immediately after all fields are initialized and when assigning after
            instantiating.
            assign: validate when field is assigned a new value (after scheme instantiating).
        """
        
        self.__func = func
        self._field = field
        self.__mode = mode

        self._field._add_validator(self)

    def _get_logger(self,
        scheme_ins: "BaseScheme",
    ):
        """Get validator level logger
        """
        return scheme_ins._logger.bind(
            validator_func=self.__func.__qualname__,
        )

    def __call__(self, 
        value: typing.Any,
        scheme_ins: "Undefined | BaseScheme" = _undefined,
        force: bool = False,
        **kwargs
    ) -> None:
        """
        :param scheme_ins: If func is instance method, provide the instance.
        :param force:
            If True, bypass mode check
        """
        if scheme_ins is _undefined:
            raise ValueError(
                "using instance method, but no scheme instance provided"
            )
        if not force:
            if self.__mode == "after":
                if not scheme_ins.__instantiated__:
                    scheme_ins.__after_field_validators__.append(self)
                    return None
            if self.__mode == "init_assign":
                if scheme_ins.__instantiated__:
                    return None
            if self.__mode == "assign":
                if not scheme_ins.__instantiated__:
                    return None
        
        logger = self._get_logger(scheme_ins=scheme_ins)

        # log extrance
        logger.debug(
            "Enter field validator",
            field_name=self._field.in_scheme_name, value=value
        )
        
        scheme_ins_copy = copy.copy(scheme_ins)
        scheme_ins_copy._set_logger(logger)
        call_as_sync(self.__func, scheme_ins_copy, value)
        return None


def field_validator(
    field: "Field[T]",
    mode: FieldValidator.ModeT = "after",
): 
    """Decorator to create a field validator.

    :param field: The field to validate.
    :param mode:

    Examples
    --------
    .. code-block:: python
        
        from blue_firmament.scheme import field_validator
        from blue_firmament.scheme import BaseScheme

        class Post(BaseScheme):
            __dal_path__ = ("post",)
            
            _id: int
            status: FieldT(str) = Field("open")

        class Comment(BaseScheme):
            __dal_path__ = ("comment",)
            
            post: FieldT[int] = Field()
            ...

            @field_validator(post)
            async def post_is_open(self, value: int):
                post_status = self.dao.select_one(Post.status, value)
                if post_status != "open":
                    raise ValueError("Post is not open")
            
        
    """
    def wrapper(func: FieldValidator.FuncT) -> FieldValidator[T]:
        """
        :param func: The function to call to validate the field.

            Supports:
            - ``def func(value: Any) -> None``
            - ``async def func(value: Any) -> None``
            - ``def func(self: BaseScheme, value: Any) -> None``
            - ``async def func(self: BaseScheme, value: Any) -> None``
        """
        return FieldValidator(field=field, func=func, mode=mode)

    return wrapper


def field_validators(
    *fields: "Field",
    mode: FieldValidator.ModeT = "after",
):
    def wrapper(func: FieldValidator.FuncT) -> typing.List[FieldValidator]:
        """
        :param func: The function to call to validate the field.

            Supports:
            - ``def func(value: Any) -> None``
            - ``async def func(value: Any) -> None``
            - ``def func(self: BaseScheme, value: Any) -> None``
            - ``async def func(self: BaseScheme, value: Any) -> None``
        """
        res = []
        for field in fields:
            res.append(FieldValidator(field=field, func=func, mode=mode))
        return res

    return wrapper


class SchemeValidator(BaseValidator):

    """Validator for a scheme.

    This validator will be added to the scheme when creating class.
    """

    class InstanceMethodFunc(typing.Protocol[BaseSchemeTV]):
        def __call__(_, self: BaseSchemeTV) -> typing.Union[
            None, typing.Coroutine[None, None, None]
        ]:
            ...

    FuncT = typing.Union[
        InstanceMethodFunc
    ]
    """Function types that SchemeValidator can accept.
    """

    def __init__(self, 
        func: FuncT,
    ) -> None:
        
        """
        :param func: The function to call to validate the scheme.
        """
        self.__func = func

    def _get_logger(self,
        scheme_ins: "BaseScheme",
    ):
        return scheme_ins._logger.bind(
            validator_func=self.__func.__qualname__,
        )

    def __call__(self, 
        scheme_ins: "BaseScheme",
        **kwargs
    ) -> None:
        
        """
        :param scheme_ins: The scheme instance to validate.
        """
        logger = self._get_logger(scheme_ins=scheme_ins)

        # log extrance
        logger.info("Enter scheme validator", 
            scheme_name=scheme_ins.__class__.__name__, value=scheme_ins.dump_to_dict()
        )

        scheme_ins_copy = copy.copy(scheme_ins)
        scheme_ins_copy._set_logger(logger)
        call_as_sync(self.__func, scheme_ins_copy)


def scheme_validator(func: SchemeValidator.FuncT) -> SchemeValidator:
    """Decorator to create a scheme validator.

    :param func: The function to call to validate the scheme.

        Supports:
        - ``def func(self: BaseScheme) -> None``
        - ``async def func(self: BaseScheme) -> None``

    """
    return SchemeValidator(func)



"""Utils for handlers."""

__all__ = [
    "retry"
]

import typing
import time

from .main import call_as_async
from ..exceptions import Retryable, MaxRetriesExceeded


def retry(
    max_: int = 1,
    default_delay: int | float = 0.2,
):
    """Retry decorated function when it raises a retryable exception.

    :param max_: The maximum number of retries
    :param default_delay: The delay between retries in seconds if
        Retryable exception does not specify a delay.

    Async and sync functions are supported and all call as async.

    Examples
    ^^^^^^^^
    ..code-block:: python
        from blue_firmament.utils.handler_ import retry
        from blue_firmament.exceptions import Retryable

        @retry()
        def your_function(a, b: int = 1):
            ... do something ...
            if some_condition:
                # recover state and prepare for retry, like refreshing token
                raise Retryable(delay=0.2)
    """
    def decorator(func: typing.Callable):
        async def wrapper(*args, **kwargs):
            for attempt in range(max_+1):
                try:
                    return await call_as_async(func, *args, **kwargs)
                except Retryable as e:
                    time.sleep(e.delay or default_delay)

            raise MaxRetriesExceeded()

        return wrapper

    return decorator